from __future__ import annotations

try:
    from typing import Self as Self
except ImportError:
    from typing_extensions import Self as Self


def get_annotations(o):
    """Access o.__annotations__ across python versions."""
    # from https://docs.python.org/3/howto/annotations.html
    # TODO: once we drop support for python 3.9, we can remove this function
    # and just use inspect.get_annotations directly
    try:
        from inspect import get_annotations

        return get_annotations(o)
    except ImportError:
        pass

    if isinstance(o, type):
        return o.__dict__.get("__annotations__", {})
    else:
        return getattr(o, "__annotations__", {})



"""Utilities for exploratory data analysis (EDA)."""

from __future__ import annotations

from mismo.eda._plot import distribution_chart as distribution_chart
from mismo.eda._plot import distribution_dashboard as distribution_dashboard



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Literal

import ibis

import mismo
from mismo import _upset, joins
from mismo.joins import HasJoinCondition
from mismo.linkage._linkage import Linkage
from mismo.linker._common import Linker, infer_task
from mismo.types._links_table import LinksTable

if TYPE_CHECKING:
    import altair as alt


class OrLinker(Linker):
    """
    A Linker that is the logical OR of multiple [mismo.HasJoinCondition].

    Physically, this is implemented as follows:
    - remove any condition overlap using [mismo.joins.remove_condition_overlap]
    - create a [LinksTable][mismo.LinksTable] for each join condition
      (which should be fast)
    - Union the [LinksTable][mismo.LinksTable]s into a single [LinksTable][mismo.LinksTable]
    """  # noqa: E501

    def __init__(
        self,
        conditions: Iterable[HasJoinCondition] | Mapping[str, HasJoinCondition],
        *,
        on_slow: Literal["error", "warn", "ignore"] = "error",
    ) -> None:
        if isinstance(conditions, Mapping):
            self._join_conditions = {
                k: joins.join_condition(v) for k, v in conditions.items()
            }
        else:
            self._join_conditions = {
                f"condition_{i}": joins.join_condition(c)
                for i, c in enumerate(conditions)
            }
        self.on_slow = on_slow

    @property
    def join_conditions(self) -> dict[str, HasJoinCondition]:
        """
        The tuple of underling HasJoinCondition objects.
        """
        return self._join_conditions

    # We explicitly do not implement __join_condition__ here,
    # because an OR join condition results in inefficient loop joins.

    def __call__(self, left: ibis.Table, right: ibis.Table) -> Linkage:
        task = infer_task(task=None, left=left, right=right)
        if left is right:
            right = right.view()
        if not self._join_conditions:
            return mismo.empty_linkage(left, right)
        conditions = [
            c.__join_condition__(left, right) for c in self._join_conditions.values()
        ]
        for c in conditions:
            joins.check_join_algorithm(left, right, c, on_slow=self.on_slow)
        conditions = joins.remove_condition_overlap(conditions)
        if task == "dedupe":
            conditions = [c & (left.record_id < right.record_id) for c in conditions]
        sub_links = [LinksTable.from_join_condition(left, right, c) for c in conditions]
        links = mismo.UnionTable(sub_links)
        return Linkage(left=left, right=right, links=links)

    def upset_chart(
        self,
        left: ibis.Table,
        right: ibis.Table,
    ) -> alt.Chart:
        import altair as alt

        blocker_names = list(self._join_conditions.keys())
        combos = _upset.combos(blocker_names)

        n_pairs_by_combo = {}
        for combo in combos:
            yes_conditions = [
                self._join_conditions[name].__join_condition__(left, right)
                for name in combo
            ]
            no_conditions = [
                ~self._join_conditions[name].__join_condition__(left, right)
                for name in blocker_names
                if name not in combo
            ]
            condition = ibis.and_(*yes_conditions, *no_conditions)
            n_pairs = mismo.join(left, right, condition).count().execute()
            n_pairs_by_combo[combo] = n_pairs

        import pandas as pd

        intersections_records = []
        for combo, n_pairs in n_pairs_by_combo.items():
            record = {"intersection_size": n_pairs}
            for name in blocker_names:
                record[name] = name in combo
            intersections_records.append(record)

        intersections_df = pd.DataFrame(intersections_records)
        chart = _upset.upset_chart(intersections_df)

        total_pairs = intersections_df.intersection_size.sum()
        title = alt.Title(
            f"{len(blocker_names)} Rules generated {total_pairs:,} total pairs",
            anchor="middle",
        )
        chart = chart.properties(title=title)
        return chart



from __future__ import annotations

import ibis
import pytest

from mismo import sets
from mismo.tests.util import assert_tables_equal


def test_document_counts(table_factory):
    addresses = [
        "12 main st",
        "34 st john ave",
        "56 oak st",
        "12 glacier st",
    ]
    t = table_factory({"address": addresses})
    tokens = t.address.re_split(r"\s+")
    result = sets.document_counts(tokens)
    expected_records = [
        ("12", 2),
        ("56", 1),
        ("main", 1),
        ("34", 1),
        ("glacier", 1),
        ("oak", 1),
        ("john", 1),
        ("st", 4),
        ("ave", 1),
    ]
    expected = ibis.memtable(expected_records, columns=["term", "n_records"])
    assert_tables_equal(result, expected)


@pytest.mark.parametrize(
    "terms, expected",
    [
        pytest.param(
            ["a", "b", "a", "b", "c", None],
            {
                "a": 2,
                "b": 2,
                "c": 1,
            },
            id="string",
        ),
        pytest.param(
            [5, 3, 3, None],
            {
                5: 1,
                3: 2,
            },
            id="int",
        ),
        pytest.param(
            [],
            {},
            id="empty",
        ),
        pytest.param(
            None,
            None,
            id="null",
        ),
    ],
)
def test_term_counts(table_factory, terms, expected):
    typ = "string" if terms and isinstance(terms[0], str) else "int64"
    inp = table_factory({"terms": [terms]}, schema={"terms": f"array<{typ}>"})
    result = sets.add_array_value_counts(inp, "terms")
    expected = table_factory(
        {"terms": [terms], "terms_counts": [expected]},
        schema={"terms": f"array<{typ}>", "terms_counts": f"map<{typ}, int64>"},
    )
    assert_tables_equal(result, expected)


def test_add_tfidf(table_factory):
    terms = [
        None,
        [],
        ["st"],
        ["12", "main", "st"],
        ["99", "main", "ave"],
        ["56", "st", "joseph", "st"],
        ["21", "glacier", "st"],
        ["12", "glacier", "st"],
    ]
    t = table_factory({"terms": terms})
    result = sets.add_tfidf(t, "terms", normalize=False)
    expected_records = [
        {"terms": ["st"], "terms_tfidf": {"st": 0.3364722366212129}},
        {
            "terms": ["12", "main", "st"],
            "terms_tfidf": {
                "12": 1.252762968495368,
                "main": 1.252762968495368,
                "st": 0.3364722366212129,
            },
        },
        {
            "terms": ["99", "main", "ave"],
            "terms_tfidf": {
                "99": 1.9459101490,
                "ave": 1.945910149,
                "main": 1.25276296,
            },
        },
        {
            "terms": ["56", "st", "joseph", "st"],
            "terms_tfidf": {
                "joseph": 1.9459101490553132,
                "56": 1.9459101490553132,
                "st": 0.6729444732424258,
            },
        },
        {
            "terms": ["21", "glacier", "st"],
            "terms_tfidf": {
                "21": 1.9459101490553132,
                "glacier": 1.252762968495368,
                "st": 0.3364722366212129,
            },
        },
        {
            "terms": ["12", "glacier", "st"],
            "terms_tfidf": {
                "12": 1.252762968495368,
                "glacier": 1.252762968495368,
                "st": 0.3364722366212129,
            },
        },
        {"terms": None, "terms_tfidf": None},
        {"terms": [], "terms_tfidf": {}},
    ]
    expected = ibis.memtable(
        expected_records,
        schema={
            "terms": "array<string>",
            "terms_tfidf": "map<string, float64>",
        },
    )
    assert_tables_equal(result, expected)



from __future__ import annotations

import ibis
from ibis.backends.duckdb import Backend as DuckDBBackend

from mismo.exceptions import UnsupportedBackendError


def explain(duckdb_expr: ibis.Expr | str, *, analyze: bool = False) -> str:
    """Get an ASCII art representation of the query plan for a given expression.

    This requires a DuckDB backend.
    """
    # we can't use a separate backend eg from ibis.duckdb.connect()
    # or it might not be able to find the tables/data referenced
    sql, con = _to_sql_and_backend(duckdb_expr)
    if analyze:
        sql = "EXPLAIN ANALYZE " + sql
    else:
        sql = "EXPLAIN " + sql
    cursor = con.raw_sql(sql)
    return cursor.fetchall()[0][1]


def _to_sql_and_backend(duckdb_expr: ibis.Expr | str) -> tuple[str, DuckDBBackend]:
    if isinstance(duckdb_expr, str):
        sql = duckdb_expr
        con = ibis.duckdb.connect()
    else:
        try:
            con = duckdb_expr._find_backend(use_default=True)
        except AttributeError:
            raise NotImplementedError("The given expression must have a backend.")
        if not isinstance(con, DuckDBBackend):
            raise UnsupportedBackendError(con)
        sql = ibis.to_sql(duckdb_expr, dialect="duckdb")
    return sql, con



from __future__ import annotations

from typing import Literal

import ibis

from mismo.linkage import _linkage
from mismo.linker import _common, _join_linker


class FullLinker(_common.Linker):
    """
    A [Linker][mismo.Linker] that yields all possible pairs.

    This will be N x M pairs for linking tasks,
    and N x (M-1) pairs for deduplication tasks.
    """

    def __init__(self, *, task: Literal["dedupe", "link"] | None = None):
        self.task = task
        self._linker = _join_linker.JoinLinker(True, on_slow="ignore", task=task)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue:
        return self._linker.__join_condition__(left, right)

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        return self._linker(left, right)


def full_linkage(
    left: ibis.Table,
    right: ibis.Table,
    *,
    task: Literal["dedupe", "link"] | None = None,
) -> _linkage.Linkage:
    """
    Create a linkage with all (M x N) possible pairs of the two tables.

    Parameters
    ----------
    left
        The left table to link.
    right
        The right table to link.
    task
        The task type, either "dedupe" or "link".
        If None, will be inferred from whether the left and right tables are the same.
    """
    return FullLinker(task=task)(left, right)


class EmptyLinker(_common.Linker):
    """A [Linker][mismo.Linker] that yields no pairs."""

    def __init__(self):
        # The task doesn't matter here, since we won't be linking anything.
        self._linker = _join_linker.JoinLinker(False, on_slow="ignore", task="link")

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue:
        return self._linker.__join_condition__(left, right)

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        return self._linker(left, right)


def empty_linkage(left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
    """
    Create a Linkage with no pairs. This is useful for testing or as a placeholder.

    Parameters
    ----------
    left
        The left table to link.
    right
        The right table to link.
    """
    return EmptyLinker()(left, right)


class UnnestLinker(_common.Linker):
    """A [Linker][mismo.Linker] that unnests a column before linking."""

    def __init__(self, column: str, *, task: Literal["dedupe", "link"] | None = None):
        self.column = column
        self.task = task
        self._linker = _join_linker.JoinLinker(self.column, task=task)

    def __call__(self, left: ibis.Table, right: ibis.Table) -> _linkage.Linkage:
        left = left.mutate(left[self.column].unnest().name(self.column))
        right = left.mutate(right[self.column].unnest().name(self.column))
        return self._linker.__call__(left, right)



from __future__ import annotations

from typing import TYPE_CHECKING

import ibis
from ibis.expr import types as ir

from mismo.joins import _conditions

if TYPE_CHECKING:
    from collections.abc import Iterable


def join(
    left: ibis.Table,
    right: ibis.Table,
    predicates: _conditions.IntoHasJoinCondition = (),
    how: str = "inner",
    *,
    lname: str = "{name}",
    rname: str = "{name}_right",
    rename_all: bool = False,
):
    """
    Ibis.join, but with enhanced condition resolution.

    This is a wrapper around `ibis.join` that

    - Allows for slightly more flexible join conditions.
    - Adds an option for renaming *all* columns with `lname` and `rname`,
      not just those that collide (the default behavior).

    Parameters
    ----------
    left
        The left table to block
    right
        The right table to block
    predicates
        What `ibis.join()` accepts as a predicate, plus some extras.

        Anything that `ibis.join()` accepts as a predicate:
        - An `ibis.BooleanValue`, such as `left.last_name == right.surname.upper()`.
        - A string, which is interpreted as the name of a column in both tables.
          eg "price" is equivalent to `left.price == right.price`.
        - A Deferred, which is used to reference a column in a table
          eg "_.price.fill_null(0)" is equivalent to `left.price.fill_null(0) == right.price.fill_null(0)`

        Plus some extra sorts of inputs:
        - A 2-tuple tuple of one of the above.
          The first element is for the left table, the second is for the right table.
          This is useful when the column names are different in the two tables,
          or require some transformation/normalization.
          For example `("last_name", _.surname.upper()")` is equivalent to
          `left.last_name == right.surname.upper()`
        - A Callable of signature (left: Table, right: Table, args) -> one of the above
    how
        Behaves the the same as in `ibis.join()`
    lname
        Behaves the the same as in `ibis.join()`
    rname
        Behaves the the same as in `ibis.join()`
    rename_all
        Should we apply `lname` and `rname` to ALL columns in the output,
        or just on the ones that collide between the two tables
        (the default, the same as in `ibis.join()`)
    """  # noqa: E501
    condition = _conditions.join_condition(predicates)
    resolved_predicate = condition.__join_condition__(left, right)
    joined = ibis.join(
        left,
        right,
        how=how,
        lname=lname,
        rname=rname,
        predicates=resolved_predicate,
    )
    if rename_all:
        joined = rename_all_joins(
            left=left, right=right, joined=joined, lname=lname, rname=rname
        )
    return joined


def rename_all_joins(
    *,
    left: ibis.Table,
    right: ibis.Table,
    joined: ibis.Table,
    lname: str = "{name}",
    rname: str = "{name}_right",
):
    """
    Adjust result of Ibis.join, always applying lname and rname, not just on conflict.
    """

    def _rename(spec: str, name: str):
        return spec.format(name=name)

    selections = []
    for col in left.columns:
        new_name = _rename(lname, col)
        if new_name in joined.columns:
            selections.append((new_name, new_name))
        else:
            assert col in joined.columns
            selections.append((new_name, col))
    for col in right.columns:
        new_name = _rename(rname, col)
        if new_name in joined.columns:
            selections.append((new_name, new_name))
        else:
            assert col in joined.columns
            selections.append((new_name, col))

    # check for dupe output columns
    by_new_name = {}
    for new_name, old_name in selections:
        if new_name not in by_new_name:
            by_new_name[new_name] = [old_name]
        else:
            by_new_name[new_name].append(old_name)
    dupes = []
    for new_name, old_names in by_new_name.items():
        if len(old_names) > 1:
            dupes.append(f"Column {new_name} is produced by {old_names}")
    if dupes:
        raise ValueError("\n".join(dupes))

    return joined.select(**{new_name: old_name for new_name, old_name in selections})


def remove_condition_overlap(
    conditions: Iterable[ir.BooleanValue | bool],
) -> list[ir.BooleanValue]:
    """
    Constrain each condition to not generate any pairs that are already created by previous conditions.
    """  # noqa: E501
    result = []
    priors = []
    for condition in conditions:
        if isinstance(condition, bool):
            condition = ibis.literal(condition)
        modified = ibis.and_(condition, *[~prior for prior in priors])
        priors.append(condition)
        result.append(modified)
    return result



"""Vector operations"""

from __future__ import annotations

from typing import Literal, TypeVar

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir


@ibis.udf.scalar.builtin(name="array_sum")
def _array_sum(a) -> float: ...


# for duckdb
@ibis.udf.scalar.builtin(name="list_dot_product")
def _array_dot_product(a, b) -> float: ...


# for duckdb
@ibis.udf.scalar.builtin(name="list_cosine_similarity")
def _array_cosine_similarity(a, b) -> float: ...


T = TypeVar("T", ir.MapValue, ir.ArrayValue)


def dot(a: T, b: T) -> ir.FloatingValue:
    """Compute the dot product of two vectors

    The vectors can either be dense vectors, represented as array<numeric>,
    or sparse vectors, represented as map<any_type, numeric>.
    Both vectors must be of the same type though.

    Parameters
    ----------
    a :
        The first vector.
    b :
        The second vector.

    Returns
    -------
    FloatingValue
        The dot product of the two vectors.

    Examples
    --------
    >>> import ibis
    >>> from mismo.vector import dot
    >>> v1 = ibis.array([1, 2])
    >>> v2 = ibis.array([4, 5])
    >>> dot(v1, v2).execute()  # 1*4 + 2*5
    14.0
    >>> m1 = ibis.map({"a": 1, "b": 2})
    >>> m2 = ibis.map({"b": 3, "c": 4})
    >>> dot(m1, m2).execute()  # 2*3
    6.0
    """
    a_vals, b_vals = _shared_vals(a, b)
    return _array_dot_product(a_vals, b_vals)


def cosine_similarity(a: T, b: T) -> ir.FloatingValue:
    """Compute the cosine similarity of two vectors

    The vectors can either be dense vectors, represented as array<numeric>,
    or sparse vectors, represented as map<any_type, numeric>.
    Both vectors must be of the same type though.

    Parameters
    ----------
    a :
        The first vector.
    b :
        The second vector.

    Returns
    -------
    FloatingValue
        The cosine similarity of the two vectors.

    Examples
    --------
    >>> import ibis
    >>> from mismo.vector import cosine_similarity

    Opposite directions:

    >>> cosine_similarity(
    ...     ibis.array([1, 1]), ibis.array([-2, -2])
    ... ).execute()  # doctest: +SKIP
    -1.0

    Orthogonal vectors:

    >>> cosine_similarity(ibis.array([1, 0]), ibis.array([0, 1])).execute()
    0.0
    """  # noqa: E501
    a_vals, b_vals = _shared_vals(a, b)
    return _array_cosine_similarity(a_vals, b_vals)


def mul(a: T, b: T) -> T:
    """Element-wise multiplication of two vectors"""
    if isinstance(a, ir.ArrayValue) and isinstance(b, ir.ArrayValue):
        return a.zip(b).map(lambda struct: struct.f1 * struct.f2)
    elif isinstance(a, ir.MapValue) and isinstance(b, ir.MapValue):
        keys = _shared_keys(a, b)
        vals = keys.map(lambda k: a[k] * b[k])
        is_null = vals.isnull()
        result = map_(keys.fill_null([]), vals.fill_null([]))
        return is_null.ifelse(ibis.null(), result)
    else:
        raise ValueError(f"Unsupported types {type(a)} and {type(b)}")


def norm(vec: T, *, metric: Literal["l1", "l2"] = "l2") -> ir.FloatingValue:
    """Compute the norm (length) of a vector.

    The vector can either be a dense vector, represented as array<numeric>,
    or a sparse vector, represented as map<any_type, numeric>.

    Parameters
    ----------
    vec :
        The vector to compute the norm of.
    metric : {"l1", "l2"}, default "l2"
        The metric to use. "l1" for Manhattan distance, "l2" for Euclidean distance.

    Returns
    -------
    The norm of the vector.

    Examples
    --------
    >>> import ibis
    >>> from mismo.vector import norm
    >>> v = ibis.array([-3, 4])
    >>> norm(v).execute()
    5.0
    >>> m = ibis.map({"a": -3, "b": 4})
    >>> norm(m, metric="l1").execute()
    7.0
    """
    if isinstance(vec, ir.ArrayValue):
        vals = vec
    elif isinstance(vec, ir.MapValue):
        vals = map_values(vec)
    else:
        raise ValueError(f"Unsupported type {type(vec)}")

    if metric == "l1":
        return _array_sum(vals.map(lambda x: x.abs()))
    elif metric == "l2":
        return _array_sum(vals.map(lambda x: x**2)).sqrt()
    else:
        raise ValueError(f"Unsupported norm {metric}")


def normalize(vec: T, *, metric: Literal["l1", "l2"] = "l2") -> T:
    """Normalize a vector to have unit length.

    The vector can either be a dense vector, represented as array<numeric>,
    or a sparse vector, represented as map<any_type, numeric>.
    The returned vector will have the same type as the input vector.

    Parameters
    ----------
    vec :
        The vector to normalize.
    metric : {"l1", "l2"}, default "l2"
        The metric to use. "l1" for Manhattan distance, "l2" for Euclidean distance.

    Returns
    -------
    ArrayValue
        The normalized vector.

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> from mismo.vector import normalize
    >>> normalize(ibis.array([1, 2])).execute()
    [0.4472135954999579, 0.8944271909999159]
    >>> normalize(ibis.array([1, 2]), metric="l1").execute()
    [0.3333333333333333, 0.6666666666666666]
    >>> normalize(ibis.map({"a": 1, "b": 2})).execute()
    {'a': 0.4472135954999579, 'b': 0.8944271909999159}
    """
    if isinstance(vec, ir.ArrayValue):
        vals = vec
    elif isinstance(vec, ir.MapValue):
        vals = map_values(vec)
    else:
        raise ValueError(f"Unsupported type {type(vec)}")

    denom = norm(vec, metric=metric)
    normed_vals = vals.map(lambda x: x / denom)

    if isinstance(vec, ir.ArrayValue):
        return normed_vals
    else:
        return map_(map_keys(vec), normed_vals)


def _shared_keys(a: ir.MapValue, b: ir.MapValue) -> ir.ArrayValue:
    regular = map_keys(a).filter(lambda k: b.contains(k))
    null = ibis.literal(None, type=dt.Array(value_type=a.type().key_type))
    return b.isnull().ifelse(null, regular)


def _shared_vals(a: T, b: T) -> tuple[ir.ArrayValue, ir.ArrayValue]:
    if isinstance(a, ir.ArrayValue) and isinstance(b, ir.ArrayValue):
        return a, b
    elif isinstance(a, ir.MapValue) and isinstance(b, ir.MapValue):
        keys = _shared_keys(a, b)
        a_vals = keys.map(lambda k: a[k])
        b_vals = keys.map(lambda k: b[k])
        return a_vals, b_vals
    else:
        raise ValueError(f"Unsupported types {type(a)} and {type(b)}")


def map_keys(m: ir.MapValue) -> ir.ArrayValue:
    """workaround for https://github.com/duckdb/duckdb/issues/11116"""
    normal = m.keys()
    if not _need_old_duckdb_workaround():
        return normal
    null = ibis.literal(None, type=dt.Array(value_type=m.type().key_type))
    return m.isnull().ifelse(null, normal)


def map_values(m: ir.MapValue) -> ir.ArrayValue:
    """workaround for https://github.com/duckdb/duckdb/issues/11116"""
    normal = m.values()
    if not _need_old_duckdb_workaround():
        return normal
    null = ibis.literal(None, type=dt.Array(value_type=m.type().value_type))
    return m.isnull().ifelse(null, normal)


def map_(keys: ir.ArrayValue, values: ir.ArrayValue) -> ir.MapValue:
    """workaround for https://github.com/duckdb/duckdb/issues/11115"""
    if not _need_old_duckdb_workaround():
        return ibis.map(keys, values)
    either_null = keys.isnull() | values.isnull()
    regular = ibis.map(keys, values)
    null = ibis.literal(
        None,
        type=dt.Map(
            key_type=keys.type().value_type, value_type=values.type().value_type
        ),
    )
    return either_null.ifelse(null, regular)


def _need_old_duckdb_workaround() -> bool:
    # Do we need workarounds for:
    # - https://github.com/duckdb/duckdb/issues/11115
    # - https://github.com/duckdb/duckdb/issues/11116
    # They were fixed in
    # https://github.com/duckdb/duckdb/releases/tag/v0.10.3
    try:
        import duckdb
    except ImportError:
        return False
    return duckdb.__version__ < "0.10.3"



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, NamedTuple

import ibis
from ibis import _
from ibis.expr import types as ir

from mismo.types._table_wrapper import TableWrapper

if TYPE_CHECKING:
    import altair as alt


class _HistSpec(NamedTuple):
    n_title: str
    chart_title: str
    chart_subtitle: str


class CountsTable(TableWrapper):
    """A table with at least an Integer column named `n`.

    There will also be variable number of other columns that act as identifiers.

    You won't create this directly, it will be returned to you
    from eg [KeyLinker.key_counts_left][mismo.KeyLinker.key_counts_left],
    [KeyLinker.key_counts_right][mismo.KeyLinker.key_counts_right],
    or [KeyLinker.pair_counts][mismo.KeyLinker.pair_counts].
    """

    n: ir.IntegerColumn
    """The column containing the count."""

    # This MUST be set in subclasses
    _KEY_COUNTS_SPEC: _HistSpec

    @functools.cache
    def n_total(self) -> int:
        """n.sum().fill_null(0), just here for convenience."""
        raw = self.n.sum().execute()
        return int(raw) if raw is not None else 0

    def chart(self, *, n_most_common: int = 50, n_least_common: int = 10) -> alt.Chart:
        return _counts_chart(
            self,
            hist_spec=self._KEY_COUNTS_SPEC,
            n_most_common=n_most_common,
            n_least_common=n_least_common,
        )


class KeyCountsTable(CountsTable):
    _KEY_COUNTS_SPEC = _HistSpec(
        "Number of Records", "Number of Records by Key", "{n_total:_} Total Records"
    )


class PairCountsTable(CountsTable):
    _PAIR_COUNTS_SPEC = _HistSpec(
        "Number of Pairs", "Number of Pairs by Key", "{n_total:_} Total Pairs"
    )


def _counts_chart(
    counts: CountsTable,
    *,
    hist_spec: _HistSpec,
    n_most_common: int,
    n_least_common: int,
):
    import altair as alt

    key_cols = [c for c in counts.columns if c != "n"]
    val = "(" + ibis.literal(", ").join(counts[c].cast(str) for c in key_cols) + ")"
    key_and_n = counts.mutate(val.name("key"), "n")
    key_and_n = key_and_n.filter(_.n > 0)
    most_common = key_and_n.order_by(_.n.desc()).head(n_most_common)
    least_common = key_and_n.order_by(_.n.asc()).head(n_least_common)
    subset = ibis.union(most_common, least_common, distinct=False).cache()
    n_keys_shown = subset.count().execute()
    n_keys_total = key_and_n.count().execute()
    key_title = "(" + ", ".join(key_cols) + ")"
    chart = (
        alt.Chart(subset)
        .properties(
            title=alt.TitleParams(
                hist_spec.chart_title,
                subtitle=[
                    hist_spec.chart_subtitle.format(n_total=counts.n_total()),
                    f"Showing the {n_keys_shown:_} most and least common keys out of {n_keys_total:_}",  # noqa: E501
                ],
                anchor="middle",
            ),
            width=alt.Step(10),
        )
        .mark_bar()
        .encode(
            alt.X("key:O", title=key_title, sort="-y"),
            alt.Y(
                "n:Q",
                title=hist_spec.n_title,
                scale=alt.Scale(type="symlog"),
            ),
            tooltip=[
                alt.Tooltip("n:Q", title=hist_spec.n_title, format=","),
                *[alt.Tooltip(col) for col in key_cols],
            ],
        )
    )
    return chart



from __future__ import annotations

from mismo.linkage._combine import difference as difference
from mismo.linkage._combine import intersect as intersect
from mismo.linkage._combine import (
    unify_links_min_intersection as unify_links_min_intersection,
)
from mismo.linkage._combine import union as union
from mismo.linkage._linkage import Linkage as Linkage
from mismo.linkage._linkage import filter_links as filter_links
from mismo.linkage._sample import sample_all_links as sample_all_links



from __future__ import annotations

import ibis
from ibis.expr import types as ir

from mismo import _util


def tokenize(text: ir.StringValue) -> ir.ArrayValue:
    """
    Split a string into tokens on whitespace.

    Examples
    --------
    >>> import ibis
    >>> from mismo.text import tokenize
    >>> tokenize(ibis.literal("  abc    def")).execute()
    ['abc', 'def']
    >>> tokenize(ibis.literal("  abc")).execute()
    ['abc']
    >>> tokenize(ibis.literal(" ")).execute()
    []
    >>> tokenize(ibis.null(str)).execute() is None
    True
    """
    stripped = text.strip()
    return (stripped == "").ifelse([], stripped.re_split(r"\s+"))


# from https://www.imperva.com/blog/fast-n-grams-extraction-and-analysis-with-sql/
def ngrams(string: ir.StringValue, n: int) -> ir.ArrayValue:
    """
    Character n-grams from a string. The order of the n-grams is not guaranteed.

    Parameters
    ----------
    string:
        The string to generate n-grams from.
    n:
        The number of characters in each n-gram.

    Returns
    -------
    An array of n-grams.

    Examples
    --------
    >>> import ibis
    >>> from mismo.text import ngrams
    >>> ngrams("abc", 2).execute()
    ['ab', 'bc']
    >>> ngrams("", 2).execute()
    []
    >>> ngrams("a", 2).execute()
    []
    >>> ngrams(None, 4).execute() is None
    True

    Order of n-grams is not guaranteed:

    >>> ngrams("abcdef", 3).execute()
    ['abc', 'def', 'bcd', 'cde']
    """
    if n < 1:
        raise ValueError("n must be greater than 0")
    string = _util.ensure_ibis(string, "string")
    pattern = "." * n
    # if you just do _re_extract_all("abcdef", "..."), you get ["abc", "def"].
    # So to get the "bcd" and the "cde", we need to offset the string
    # by one and two (in general up to n-1) characters.
    first, *rest = [_re_extract_all(string[i:], pattern) for i in range(0, n)]
    return string.isnull().ifelse(ibis.null("array<string>"), first.concat(*rest))


@ibis.udf.scalar.builtin(
    name="regexp_extract_all", signature=(("string", "string"), "array<string>")
)
def _re_extract_all(string, pattern): ...



from __future__ import annotations

import pytest

from mismo.linker import IDLinker
from mismo.tests.util import assert_tables_equal


@pytest.fixture
def a(table_factory):
    return table_factory(
        {
            "record_id": [1, 2, 3, 4],
            "label": [9, 9, None, 8],
        }
    )


@pytest.fixture
def b(table_factory):
    return table_factory(
        {
            "record_id": [11, 12, 13, 14],
            "label": [7, None, 8, 8],
        }
    )


@pytest.mark.xfail(raises=NotImplementedError)
@pytest.mark.parametrize(
    "when_not_equal_val, when_null_val, expected_data",
    [
        (
            "nonmatch",
            "nonmatch",
            [],
        ),
        (
            "nonmatch",
            "indefinite",
            [
                (3, 11),
                (3, 12),
                (3, 13),
                (3, 14),
                (1, 12),
                (2, 12),
                (4, 12),
            ],
        ),
        (
            "indefinite",
            "nonmatch",
            [
                (1, 11),
                (1, 13),
                (1, 14),
                (2, 11),
                (2, 13),
                (2, 14),
                (4, 11),
            ],
        ),
        (
            "indefinite",
            "indefinite",
            [
                (1, 11),
                (1, 12),
                (1, 13),
                (1, 14),
                (2, 11),
                (2, 12),
                (2, 13),
                (2, 14),
                (3, 11),
                (3, 12),
                (3, 13),
                (3, 14),
                (4, 11),
                (4, 12),
            ],
        ),
    ],
)
def test_indefinite_linkage(
    table_factory, a, b, when_not_equal_val, when_null_val, expected_data
):
    linker = IDLinker(
        "label", when_not_equal=when_not_equal_val, when_null=when_null_val
    )
    linkage = linker.indefinite_linkage(a, b)
    actual = linkage.links.select("record_id_l", "record_id_r")
    expected = table_factory(
        expected_data,
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual)


@pytest.mark.parametrize(
    "when_not_equal_val, when_null_val",
    [
        ("nonmatch", "nonmatch"),
        ("nonmatch", "indefinite"),
        ("indefinite", "nonmatch"),
        ("indefinite", "indefinite"),
    ],
)
def test_match_linkage(table_factory, a, b, when_not_equal_val, when_null_val):
    linker = IDLinker(
        "label", when_not_equal=when_not_equal_val, when_null=when_null_val
    )
    linkage = linker.match_linkage(a, b)
    actual = linkage.links.select("record_id_l", "record_id_r")
    expected = table_factory(
        [
            (4, 13),
            (4, 14),
        ],
        schema=actual.schema(),
    )
    assert_tables_equal(expected, actual)



from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, Iterable, Mapping

import ibis
from ibis import _
from ibis.expr import types as ir
from IPython.display import display
import ipywidgets
import rich
import solara

from mismo import _util
from mismo._datasets import Datasets
from mismo.cluster._connected_components import connected_components

if TYPE_CHECKING:
    import ipycytoscape  # type: ignore


def cytoscape_widget(
    tables: Datasets | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table],
    links: ir.Table,
) -> ipycytoscape.CytoscapeWidget:
    """Make a ipycytoscape.CytoscapeWidget that shows records and links.

    This shows ALL the supplied records and links, so be careful,
    you probably want to filter them down first.

    Parameters
    ----------
    tables :
        Table(s) of records with at least the column `record_id`.
    links :
        A table of edges with at least columns
        (record_id_l, record_id_r) and optionally other columns.
        The column `width` is used to set the width of the edges.
        If not given, it is determined from the column `odds`, if
        present, or set to 5 otherwise.
        The column `opacity` is used to set the opacity of the edges.
        If not given, it is set to 0.5.
    """
    with _util.optional_import("ipycytoscape"):
        import ipycytoscape  # type: ignore

    ds = Datasets(tables)
    links = _filter_links(links, ds)
    graph = {"nodes": _nodes_to_json(ds), "edges": _edges_to_json(links)}
    cyto = ipycytoscape.CytoscapeWidget(graph, layout={"name": "fcose"})
    style = [
        *cyto.get_style(),
        {
            "selector": "node",
            "css": {
                "label": "data(label)",
                "font-size": 8,
                "color": "data(color)",
                "width": 15,
                "height": 15,
            },
        },
        {
            "selector": "edge",
            "css": {
                "curve-style": "straight",
                "width": "data(width)",
                "opacity": "data(opacity)",
            },
        },
    ]
    cyto.set_style(style)
    return cyto


def _filter_links(links: ir.Table, ds: Datasets) -> ir.Table:
    if "record_id_l" not in links.columns:
        raise ValueError("links must have a record_id_l column")
    if "record_id_r" not in links.columns:
        raise ValueError("links must have a record_id_r column")
    links = links.filter(
        _.record_id_l.isin(ds.all_record_ids()),
        _.record_id_r.isin(ds.all_record_ids()),
    )
    return links


def _nodes_to_json(ds: Datasets) -> list[dict[str, Any]]:
    colors = ["blue", "red", "green"]
    cmap = dict(zip(ds.names, colors[: len(ds)]))

    def f(name: str, t: ir.Table) -> ir.Table:
        m = {
            "dataset": ibis.literal(name),
            "id": _.record_id.cast(str),
        }
        if "label" not in t.columns:
            m["label"] = name + ":" + _.record_id.cast(str)
        if "color" not in t.columns:
            m["color"] = ibis.literal(cmap[name])
        return t.mutate(m)

    return _to_json(*ds.map(f))


def _edges_to_json(links: ir.Table) -> list[dict[str, Any]]:
    def _ensure_has_width(links: ir.Table) -> ir.Table:
        if "width" in links.columns:
            return links
        if "odds" not in links.columns:
            return links.mutate(width=5)
        log_odds = _.odds.log10()
        log_odds_fraction = log_odds / log_odds.max()
        width = 10 * log_odds_fraction
        return links.mutate(width=width)

    def _ensure_has_opacity(links: ir.Table) -> ir.Table:
        if "opacity" in links.columns:
            return links
        return links.mutate(opacity=0.5)

    if "source" in links.columns:
        raise ValueError("links must not have a source column")
    if "target" in links.columns:
        raise ValueError("links must not have a target column")
    links = links.mutate(source="record_id_l", target="record_id_r")
    links = _ensure_has_width(links)
    links = _ensure_has_opacity(links)
    return _to_json(links)


def _to_json(*tables: ir.Table) -> list[dict[str, Any]]:
    records = []
    for t in tables:
        # include default_handler to avoid https://stackoverflow.com/a/60492211/5156887
        json_str = t.to_pandas().to_json(
            orient="records", default_handler=str, date_format="iso"
        )
        records.extend(json.loads(json_str))
    return records


def _render_to_html(obj) -> str:
    console = rich.console.Console(record=True)
    with console.capture():
        console.print(obj)
    template = """
    <pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace; line-height: 1.2em">
        <code style="font-family:inherit">{code}</code>
    </pre>
    """  # noqa: E501
    return console.export_html(code_format=template, inline_styles=True)


@solara.component
def cluster_dashboard(
    ds: Datasets | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table],
    links: ir.Table,
) -> solara.Column:
    ds = Datasets(ds)

    def get_output():
        out = ipywidgets.Output()
        out.append_display_data(ipywidgets.HTML("Select a node or edge..."))
        return out

    info = solara.use_memo(get_output)

    def make_cyto() -> tuple[Any, dict[Any, dict]]:
        cyto = cytoscape_widget(ds, links)
        lookup = {r["record_id"]: r for r in _nodes_to_json(ds)}

        def on_record(node: dict[str, Any]):
            info.clear_output()
            html_widget = ipywidgets.HTML(_render_to_html(node["data"]))
            with info:
                display(html_widget)

        def on_edge(edge: dict[str, Any]):
            s = edge["data"]["record_id_l"]
            t = edge["data"]["record_id_r"]
            record_l = lookup[s]
            record_r = lookup[t]
            box = ipywidgets.HBox(
                [
                    ipywidgets.HTML(_render_to_html(record_l)),
                    ipywidgets.HTML(_render_to_html(record_r)),
                ]
            )
            info.clear_output()
            with info:
                display(box)

        cyto.on("node", "click", on_record)
        cyto.on("edge", "click", on_edge)
        return cyto

    cyto = solara.use_memo(make_cyto, [ds, links])

    return solara.Column([cyto, info])


@solara.component
def clusters_dashboard(
    tables: Datasets | ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table],
    links: ir.Table,
) -> solara.Column:
    def get_ds():
        ds = Datasets(tables)
        li = _filter_links(links, ds)
        ds = connected_components(records=ds, links=li)
        ds = ds.cache()
        return ds

    ds = solara.use_memo(get_ds, [tables, links])

    def get_components():
        return (
            ds.unioned()
            .select("component")
            .distinct()
            .order_by("component")
            .component.execute()
            .to_list()
        )

    all_components = solara.use_memo(get_components, [ds])

    component = solara.use_reactive(all_components[0] if len(all_components) else None)
    component_selector = solara.Select(
        "Component", values=all_components, value=component
    )

    def get_subgraph():
        return ds.map(lambda name, t: t.filter(_.component == component.value))

    subgraph = solara.use_memo(get_subgraph, [ds, component.value])
    return solara.Column([component_selector, cluster_dashboard(subgraph, links)])



from __future__ import annotations

import ibis
import pytest

from mismo.lib import geo


@pytest.mark.parametrize(
    "inp,expected",
    [
        ("GENERAL DELIVERY", None),
        ("BARRACKS ST UNIT 2", None),
        (
            "4602 CR 673",
            {
                "AddressNumber": "4602",
                "StreetNamePreDirectional": "",
                "StreetName": "CR 673",
                "StreetNamePostType": "",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "6473 FM 1798",
            {
                "AddressNumber": "6473",
                "StreetNamePreDirectional": "",
                "StreetName": "FM 1798",
                "StreetNamePostType": "",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "1 1ST",
            {
                "AddressNumber": "1",
                "StreetName": "1ST",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "123 MAIN",
            {
                "AddressNumber": "123",
                "StreetName": "MAIN",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "123 S MAIN WEST",
            {
                "AddressNumber": "123",
                "StreetName": "MAIN",
                "StreetNamePostDirectional": "WEST",
                "StreetNamePostType": "",
                "StreetNamePreDirectional": "S",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "1 1ST ST",
            {
                "AddressNumber": "1",
                "StreetName": "1ST",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "ST",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "7100 RIDGE MANOR LN",
            {
                "AddressNumber": "7100",
                "StreetNamePreDirectional": "",
                "StreetName": "RIDGE MANOR",
                "StreetNamePostType": "LN",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # Mixed case is also detected
        (
            "7100 Ridge Manor ln",
            {
                "AddressNumber": "7100",
                "StreetNamePreDirectional": "",
                "StreetName": "Ridge Manor",
                "StreetNamePostType": "ln",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "1101B ODIN ST",
            {
                "AddressNumber": "1101B",
                "StreetNamePreDirectional": "",
                "StreetName": "ODIN",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "1385 JUNEAU AVE",
            {
                "AddressNumber": "1385",
                "StreetNamePreDirectional": "",
                "StreetName": "JUNEAU",
                "StreetNamePostType": "AVE",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "PO BOX 651",
            {
                "AddressNumber": "",
                "StreetNamePreDirectional": "",
                "StreetName": "",
                "StreetNamePostType": "",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "651",
            },
        ),
        (
            "p.o. Box 651",
            {
                "AddressNumber": "",
                "StreetNamePreDirectional": "",
                "StreetName": "",
                "StreetNamePostType": "",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "651",
            },
        ),
        (
            "604 FIFTH ST",
            {
                "AddressNumber": "604",
                "StreetNamePreDirectional": "",
                "StreetName": "FIFTH",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "604 E  FIFTH ST",
            {
                "AddressNumber": "604",
                "StreetNamePreDirectional": "E",
                "StreetName": "FIFTH",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "604 E  5TH ST",
            {
                "AddressNumber": "604",
                "StreetNamePreDirectional": "E",
                "StreetName": "5TH",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "44755 STERLING HWY",
            {
                "AddressNumber": "44755",
                "StreetNamePreDirectional": "",
                "StreetName": "STERLING",
                "StreetNamePostType": "HWY",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "100 MC CARREY ST",
            {
                "AddressNumber": "100",
                "StreetNamePreDirectional": "",
                "StreetName": "MC CARREY",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # periods after abbreviations are ignored
        (
            "100 MC CARREY ST.",
            {
                "AddressNumber": "100",
                "StreetNamePreDirectional": "",
                "StreetName": "MC CARREY",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "100 MC CARREY ST. W",
            {
                "AddressNumber": "100",
                "StreetNamePreDirectional": "",
                "StreetName": "MC CARREY",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "W",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "100 MC CARREY ST. W.",
            {
                "AddressNumber": "100",
                "StreetNamePreDirectional": "",
                "StreetName": "MC CARREY",
                "StreetNamePostType": "ST",
                "StreetNamePostDirectional": "W",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # But a perdiod in the middle of a street name is not ignored
        (
            "100 St. Johann Loop",
            {
                "AddressNumber": "100",
                "StreetNamePreDirectional": "",
                "StreetName": "St. Johann",
                "StreetNamePostType": "Loop",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # I kid you not this is a legal street type
        (
            "1015 OTTER RUN",
            {
                "AddressNumber": "1015",
                "StreetNamePreDirectional": "",
                "StreetName": "OTTER",
                "StreetNamePostType": "RUN",
                "StreetNamePostDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # 3 token long street name
        (
            "9950 STEPHEN RICHARDS MEMORIAL DR",
            {
                "AddressNumber": "9950",
                "StreetName": "STEPHEN RICHARDS MEMORIAL",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "DR",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # 4 token long street name
        (
            "2370 S RUE DE LA PAIX LOOP",
            {
                "AddressNumber": "2370",
                "StreetName": "RUE DE LA PAIX",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "LOOP",
                "StreetNamePreDirectional": "S",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # zero instead of O
        (
            "P 0 BOX 33",
            {
                "AddressNumber": "",
                "StreetName": "",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "33",
            },
        ),
        # apostrophe in street name
        (
            "4421 O'MALLEY RD",
            {
                "AddressNumber": "4421",
                "StreetName": "O'MALLEY",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "RD",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # hyphen in street name
        (
            "46465 HOLT-LAMPLIGHT RD",
            {
                "AddressNumber": "46465",
                "StreetName": "HOLT-LAMPLIGHT",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "RD",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        (
            "HC 1 BOX 106",
            {
                "AddressNumber": "",
                "StreetName": "",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "HC",
                "USPSBoxGroupID": "1",
                "USPSBoxID": "106",
            },
        ),
        (
            "MILE 15 ALYESKA HWY",
            {
                "AddressNumber": "15",
                "StreetName": "ALYESKA",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "HWY",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # with decimal 1
        (
            "MILE 1.8 DYEA RD",
            {
                "AddressNumber": "1.8",
                "StreetName": "DYEA",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "RD",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
        # with decimal 2
        (
            "MILE 15.3 ALYESKA HWY",
            {
                "AddressNumber": "15.3",
                "StreetName": "ALYESKA",
                "StreetNamePostDirectional": "",
                "StreetNamePostType": "HWY",
                "StreetNamePreDirectional": "",
                "USPSBoxGroupType": "",
                "USPSBoxGroupID": "",
                "USPSBoxID": "",
            },
        ),
    ],
)
def test_parse_street1_re(inp, expected):
    result = geo.parse_street1_re(ibis.literal(inp)).execute()
    assert expected == result



from __future__ import annotations

import ibis
from ibis.expr import types as ir
import pandas as pd
import pytest

from mismo.cluster import connected_components
from mismo.tests.util import get_clusters


@pytest.fixture(params=["component", "cluster"])
def label_as(request):
    return request.param


@pytest.mark.parametrize(
    "edges_list, edges_dtype, expected_clusters",
    [
        pytest.param(
            [
                (0, 10),
                (1, 10),
                (1, 11),
                (2, 11),
                (2, 12),
                (9, 20),
            ],
            "int64",
            {
                frozenset({0, 1, 2, 10, 11, 12}),
                frozenset({9, 20}),
            },
            id="linear chain and singleton",
        ),
        pytest.param(
            [
                (0, 10),
                (0, 11),
                (0, 12),
                (0, 13),
                (9, 20),
            ],
            "int64",
            {
                frozenset({0, 10, 11, 12, 13}),
                frozenset({9, 20}),
            },
            id="hub and singleton",
        ),
        pytest.param(
            [
                ("a", "b"),
                ("a", "c"),
                ("a", "d"),
                ("a", "e"),
                ("x", "y"),
            ],
            "string",
            {
                frozenset({"a", "b", "c", "d", "e"}),
                frozenset({"x", "y"}),
            },
            id="hub and singleton (string keys)",
        ),
        pytest.param(
            [
                (("a", 0), ("b", 10)),
                (("a", 0), ("b", 11)),
                (("a", 0), ("b", 12)),
                (("a", 0), ("b", 13)),
                (("a", 9), ("b", 20)),
            ],
            "struct<dataset: string, record_id: int64>",
            {
                frozenset({("a", 0), ("b", 10), ("b", 11), ("b", 12), ("b", 13)}),
                frozenset({("a", 9), ("b", 20)}),
            },
            # this could be useful if different datasets have the same record_id
            id="struct keys",
        ),
        pytest.param(
            [],
            "int64",
            set(),
            id="empty",
        ),
        pytest.param(
            [(42, 42)],
            "int64",
            {frozenset({42})},
            id="single self-loop",
        ),
        pytest.param(
            [(0, 1)],
            "int64",
            {frozenset({0, 1})},
            id="single edge",
        ),
    ],
)
def test_cc_only_edges(
    table_factory, edges_list, label_as, edges_dtype, expected_clusters
):
    """returns a mapping table record_id -> component"""
    schema = {"record_id_l": edges_dtype, "record_id_r": edges_dtype}
    edges_df = pd.DataFrame(edges_list, columns=["record_id_l", "record_id_r"])
    links = table_factory(edges_df, schema=schema).cast(schema)
    labels = connected_components(links=links, label_as=label_as)
    clusters = _labels_to_clusters(labels, label_as)
    assert clusters == expected_clusters


def test_cc_single_records(table_factory, label_as):
    """augments the passed records with a component column"""
    links = table_factory([(0, 1), (1, 2)], columns=["record_id_l", "record_id_r"])
    nodes = table_factory({"record_id": [0, 1, 2, 3]})
    labeled = connected_components(links=links, records=nodes, label_as=label_as)
    clusters = _labels_to_clusters(labeled, label_as)
    assert clusters == {frozenset({0, 1, 2}), frozenset({3})}


def test_cc_no_links_but_records(table_factory, label_as):
    """If there are no links, each record should be its own cluster."""
    link_schema = {"record_id_l": "int64", "record_id_r": "int64"}
    link_df = pd.DataFrame({"record_id_l": [], "record_id_r": []})
    links = table_factory(link_df, schema=link_schema)
    nodes = table_factory({"record_id": [0, 1, 2]})
    labeled = connected_components(links=links, records=nodes, label_as=label_as)
    clusters = _labels_to_clusters(labeled, label_as)
    assert clusters == {frozenset({0}), frozenset({1}), frozenset({2})}


def test_cc_multi_records(table_factory, label_as):
    # multiple input record tables
    links = table_factory([(0, 1), (1, 2)], columns=["record_id_l", "record_id_r"])
    nodes1 = table_factory({"record_id": [0, 1]})
    nodes2 = table_factory({"record_id": [2, 3]})
    l1, l2 = connected_components(
        links=links, records=(nodes1, nodes2), label_as=label_as
    )
    assert set(l1.record_id.execute()) == {0, 1}
    assert set(l2.record_id.execute()) == {2, 3}
    clusters = _labels_to_clusters(ibis.union(l1, l2), label_as)
    assert clusters == {frozenset({0, 1, 2}), frozenset({3})}


def test_cc_max_iterations(table_factory):
    """If we don't give it adequate iterations, it should not give the right result."""
    links = table_factory([(0, 1), (1, 2)], columns=["record_id_l", "record_id_r"])
    labels = connected_components(links=links, max_iter=1)
    clusters = _labels_to_clusters(labels)
    assert clusters != {frozenset({0, 1, 2})}


def _labels_to_clusters(
    labels: ir.Table, label_as: str = "component"
) -> set[frozenset[int]]:
    return get_clusters(labels[label_as], label=labels.record_id)



from __future__ import annotations

import ibis
import numpy as np
import pytest

from mismo.sets import jaccard


@pytest.mark.parametrize(
    "a, b, expected",
    [
        ([], [], 0),
        ([], [1], 0),
        ([1], [], 0),
        ([1], [1], 1),
        ([1], [1, 2], 1 / 2),
        (None, [1], np.nan),
        (None, None, np.nan),
        ([1, 2, 3, 4], [3, 4, 5, 6, 7, 8], 1 / 4),
    ],
)
def test_jaccard(a, b, expected):
    a = ibis.literal(a, type="array<int64>")
    b = ibis.literal(b, type="array<int64>")
    result = jaccard(a, b).execute()
    assert result == pytest.approx(expected, abs=0, nan_ok=True)



from __future__ import annotations

import ibis


class TableWrapper(ibis.Table):
    """A wrapper around an ibis Table that allows you to access its attributes.

    This allows the user to pretend that they are working with an ibis Table:

    - `isinstance(obj, ibis.Table)` is True
    - users can call all normal methods `t.distinct()`
    - can access columns, eg `t.column_name` or `t["column_name"]`

    but it also allows use as developers to extend the functionality of the table,
    eg adding custom methods or attributes.
    """

    _t: ibis.Table
    """The underlying Table that this is a proxy for."""

    def __init__(self, t: ibis.Table) -> None:
        object.__setattr__(self, "_t", t)

    def __getattr__(self, key: str):
        try:
            object.__getattr__(self, key)
        except AttributeError:
            pass
        return getattr(self._t, key)



from __future__ import annotations

from typing import Protocol

from ibis.expr import types as ir


class PRecipe(Protocol):
    """Preps, blocks, and compares records."""

    def prepare(self, t: ir.Table) -> ir.Table:
        """Prepare the data for blocking and comparison."""

    def block(self, t1: ir.Table, t2: ir.Table, **kwargs) -> ir.Table:
        """Block tables into record pairs."""

    def compare(self, t: ir.Table) -> ir.Table:
        """Perform comparisons on the blocked pairs."""



from __future__ import annotations

import logging
from typing import Callable, Generic, Iterable, TypeVar

logger = logging.getLogger(__name__)

Ret = TypeVar("Ret")
F = TypeVar("F", bound=Callable[..., Ret])


class Registry(Generic[F, Ret]):
    """A registry for different implementations, similar to functools.singledispatch.

    An implementation can be registered with the `register` method.

    When the registry is called, it will try each registered implementation in order.
    Each implementation is a callable that can either:
    - Return the sentinel value `NotImplemented` if it cannot handle the input,
      signaling that the next implementation should be tried.
    - Return anything else, which will be returned as the result.

    Examples
    --------
    >>> from mismo._registry import Registry
    >>> import json
    >>> import datetime
    >>> to_json = Registry[Callable[..., str], str]()
    >>> @to_json.register
    ... def _default(obj: object) -> str:
    ...     try:
    ...         return json.dumps(obj)
    ...     except TypeError:
    ...         # If the object is not JSON serializable,
    ...         # fall back to one of the registered implementations.
    ...         return NotImplemented
    >>> @to_json.register
    ... def _datetime(dt: datetime.datetime) -> str:
    ...     return dt.isoformat()
    >>> @to_json.register
    ... def _date(date: datetime.date) -> str:
    ...     return date.isoformat()
    >>> to_json(datetime.datetime(2023, 1, 1, 12, 0, 0))
    '2023-01-01T12:00:00'
    >>> to_json(datetime.date(2023, 1, 1))
    '2023-01-01'
    >>> to_json({"key": 4})
    '{"key": 4}'
    """

    def __init__(
        self,
        implementations: Iterable[F] = (),
    ) -> None:
        self.implementations = tuple(implementations)
        """Mutable, so users can modify as needed."""

    def register(self, implementation: F) -> F:
        """
        Register (after existing implementations) a new implementation of the Registry.
        """
        self.implementations = (*self.implementations, implementation)
        return implementation

    def __call__(self, *args, **kwargs) -> Ret:
        """
        Combine multiple Linkages by finding the first implementation that matches.
        """
        for implementation in self.implementations:
            result = implementation(*args, **kwargs)
            if result is NotImplemented:
                continue
            else:
                return result
        raise NotImplementedError



from __future__ import annotations

from typing import Protocol, runtime_checkable

from ibis.expr import types as ir


@runtime_checkable
class PComparer(Protocol):
    """A Callable that adds column(s) of features to a table of record pairs."""

    def __call__(self, pairs: ir.Table, **kwargs) -> ir.Table:
        """Add column(s) of features to a table of record pairs.

        For example, add a match score to each record pair, modify a score from
        a previous PComparer, or similar.

        Implementers *must* expect to be called with a table of record pairs.
        Columns suffixed with "_l" come from the left table, columns suffixed
        with "_r" come from the right table, and columns with neither suffix
        are features of the pair itself (eg from a different PComparer).
        """



from __future__ import annotations

import ibis
import pytest

from mismo.playdata import (
    load_febrl1,
    load_febrl2,
    load_febrl3,
    load_patents,
    load_rldata500,
    load_rldata10000,
)


@pytest.mark.parametrize(
    "load_func, expected_count, expected_link_count",
    [
        (load_febrl1, 1000, 500),
        (load_febrl2, 5000, 1934),
        (load_febrl3, 5000, 6538),
    ],
)
def test_load_febrl_smoketest(load_func, expected_count, expected_link_count):
    data, links = load_func()
    assert data.count().execute() == expected_count
    assert links.count().execute() == expected_link_count
    repr(data)


def test_load_patents_smoketest():
    dataset = load_patents()
    assert dataset.count().execute() == 2379
    repr(dataset)


class TestRLData:
    EXPECTED_SCHEMA = {
        "record_id": "int64",
        "label_true": "int64",
        "fname_c1": "string",
        "fname_c2": "string",
        "lname_c1": "string",
        "lname_c2": "string",
        "by": "int64",
        "bm": "int64",
        "bd": "int64",
    }

    def test_load_rldata500(self):
        dataset = load_rldata500()

        assert dataset.schema().equals(ibis.schema(TestRLData.EXPECTED_SCHEMA))
        assert dataset.count().execute() == 500
        assert dataset["label_true"].nunique().execute() == 450  # 50 duplicates

        # Proportion of clusters where surname is not unique.
        # Should be around 3.5%, i.e. around one third of the
        # 11% of clusters with more than one record.
        name_variation_rate = 1 - (
            dataset.aggregate(
                by="label_true", unique_surname=dataset["lname_c1"].nunique() == 1
            )["unique_surname"].mean()
        )
        assert name_variation_rate.execute() == 0.03555555555555556

    def test_load_rldata10000(self):
        dataset = load_rldata10000()

        assert dataset.schema().equals(ibis.schema(TestRLData.EXPECTED_SCHEMA))
        assert dataset.count().execute() == 10000
        assert dataset["label_true"].nunique().execute() == 9000  # 9000 duplicates

        name_variation_rate = 1 - (
            dataset.aggregate(
                by="label_true", unique_surname=dataset["lname_c1"].nunique() == 1
            )["unique_surname"].mean()
        )
        assert name_variation_rate.execute() == 0.030444444444444496



from __future__ import annotations

import ibis
from ibis.expr import datatypes as dt
from ibis.expr import types as ir

_ARRAY_AGGS = {}


def _get_array_agg(array: ir.ArrayValue, name: str, *, out_type=None) -> ir.Column:
    t = array.type()
    if not isinstance(array, ir.ArrayValue):
        raise ValueError(f"Expected an array, got {t}")

    if out_type is None:
        out_type = t.value_type()

    key = (t, name)
    if key not in _ARRAY_AGGS:

        @ibis.udf.scalar.builtin(name=name, signature=((t,), out_type))
        def f(array): ...

        _ARRAY_AGGS[key] = f

    return _ARRAY_AGGS[key](array)


def array_min(array: ir.ArrayValue) -> ir.NumericValue:
    """Get the minimum value of an array."""
    return _get_array_agg(array, "list_min")


def array_sum(array: ir.ArrayValue) -> ir.NumericValue:
    """Get the sum of all values of an array."""
    return _get_array_agg(array, "list_sum")


def array_max(array: ir.ArrayValue) -> ir.NumericValue:
    """Get the maximum value of an array."""
    return _get_array_agg(array, "list_max")


# @ibis.udf.scalar.builtin(name="list_avg")
def array_mean(array: ir.ArrayValue) -> ir.FloatingValue:
    """Get the mean value of an array."""
    return _get_array_agg(array, "list_avg", out_type=dt.float64)


# @ibis.udf.scalar.builtin(name="list_median")
def array_median(array: ir.ArrayValue) -> ir.FloatingValue:
    """Get the median value of an array."""
    return _get_array_agg(array, "list_median", out_type=dt.float64)


@ibis.udf.scalar.builtin(name="list_bool_or")
def array_any(array) -> bool:
    """Return True if any elements in the array are True, False otherwise.

    NULL values are ignored.
    If there are no non-NULL values, returns NULL.
    """


@ibis.udf.scalar.builtin(name="list_bool_and")
def array_all(array) -> bool:
    """Return True if all elements in the array are True, False otherwise.

    NULL values are ignored.
    If there are no non-NULL values, returns NULL.
    """



"""Functionalities for set and Bag-of-Words processing and analysis."""

from __future__ import annotations

from mismo.sets._compare import jaccard as jaccard
from mismo.sets._tfidf import add_array_value_counts as add_array_value_counts
from mismo.sets._tfidf import add_tfidf as add_tfidf
from mismo.sets._tfidf import document_counts as document_counts
from mismo.sets._tfidf import rare_terms as rare_terms
from mismo.sets._tfidf import term_idf as term_idf



from __future__ import annotations

import ibis
from ibis.common.exceptions import IntegrityError
import pytest

from mismo import joins


def test_join_ensure_named():
    a = ibis.table({"a": "int64", "b": "string"}, name="t")
    b = ibis.table({"a": "int64", "c": "string"}, name="u")
    x = ibis.table({"y": "int64", "z": "string"}, name="u")

    def _schema(t):
        return {k: str(v) for k, v in t.schema().items()}

    assert {
        "a": "int64",
        "b": "string",
        "a_right": "int64",
        "c_right": "string",
    } == _schema(joins.join(a, b, "a", rename_all=True))
    assert {
        "a": "int64",
        "b": "string",
        "a_right": "int64",
        "c_right": "string",
    } == _schema(joins.join(a, b, a.a > b.a, rename_all=True))
    assert {
        "a": "int64",
        "b": "string",
        "y_right": "int64",
        "z_right": "string",
    } == _schema(joins.join(a, x, True, rename_all=True))

    assert {
        "a_l": "int64",
        "b_l": "string",
        "a_right": "int64",
        "c_right": "string",
    } == _schema(joins.join(a, b, a.a > b.a, lname="{name}_l", rename_all=True))
    assert {
        "a_l": "int64",
        "b_l": "string",
        "a": "int64",
        "c": "string",
    } == _schema(
        joins.join(a, b, a.a > b.a, lname="{name}_l", rname="{name}", rename_all=True)
    )

    with pytest.raises(IntegrityError):
        joins.join(a, b, a.a > b.a, lname="{name}_x", rname="{name}_x", rename_all=True)



from __future__ import annotations

import importlib.metadata
import warnings

from mismo import arrays as arrays
from mismo import cluster as cluster
from mismo import compare as compare
from mismo import eda as eda
from mismo import exceptions as exceptions
from mismo import fs as fs
from mismo import joins as joins
from mismo import lib as lib
from mismo import linkage as linkage
from mismo import linker as linker
from mismo import playdata as playdata
from mismo import sets as sets
from mismo import text as text
from mismo import tf as tf
from mismo import types as types
from mismo import vector as vector
from mismo._counts_table import CountsTable as CountsTable
from mismo._datasets import Datasets as Datasets
from mismo._explain import explain as explain
from mismo._n_naive import n_naive_comparisons as n_naive_comparisons
from mismo._recipe import PRecipe as PRecipe
from mismo.joins import HasJoinCondition as HasJoinCondition
from mismo.joins import IntoHasJoinCondition as IntoHasJoinCondition
from mismo.joins import join as join
from mismo.joins import join_condition as join_condition
from mismo.joins import left as left
from mismo.joins import right as right
from mismo.linkage import Linkage as Linkage
from mismo.linker import EmptyLinker as EmptyLinker
from mismo.linker import FullLinker as FullLinker
from mismo.linker import IDLinker as IDLinker
from mismo.linker import JoinLinker as JoinLinker
from mismo.linker import KeyLinker as KeyLinker
from mismo.linker import Linker as Linker
from mismo.linker import OrLinker as OrLinker
from mismo.linker import empty_linkage as empty_linkage
from mismo.linker import full_linkage as full_linkage
from mismo.types import Diff as Diff
from mismo.types import DiffStats as DiffStats
from mismo.types import LinkCountsTable as LinkCountsTable
from mismo.types import LinkedTable as LinkedTable
from mismo.types import LinksTable as LinksTable
from mismo.types import UnionTable as UnionTable
from mismo.types import Updates as Updates

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError as e:
    warnings.warn(f"Could not determine version of {__name__}\n{e!s}", stacklevel=2)
    __version__ = "unknown"



from __future__ import annotations

from typing import Any, Iterable, Literal

import ibis
from ibis.expr import types as ir

from mismo import _resolve
from mismo.linkage._linkage import Linkage, LinksTable
from mismo.linker._key_linker import KeyLinker


class IDLinker:
    """
    Helper for linking unambiguous IDs, such as social security numbers, usernames, etc.

    Where the IDs match, we treat them as a match.

    Where the IDs are uneqal, we can either treat them as a non-match or as a maybe.
    Similar for when one or both of the IDs are null,
    we can either treat them as a non-match or as a maybe.

    For example, say you are linking user accounts.
    If the emails match, then you know they are the same user.
    If they don't match, then they MIGHT be the same user, but you don't know,
    and will have to look at e.g. the IP address, the name on the account, etc.
    So this would be a `when_not_equal="indefinite"` situation.

    On the other hand, say you are linking records of people
    with social security numbers.
    If the SSN match, then you know they are the same person.
    If the SSN don't match, then you know if they are not the same person.
    So this would be a `when_not_equal="nonmatch"` situation.

    Similar for when one or both of the IDs are null.
    Depending on the situation,
    this can either signify that the records are a non-match,
    or that we don't know, and we will leave that decision to other
    parts of the linkage pipeline.

    See https://github.com/NickCrews/mismo/issues/77 for more information.
    """  # noqa: E501

    def __init__(
        self,
        labels: Iterable[str | ibis.Deferred | Any],
        *,
        when_null: Literal["nonmatch", "indefinite"],
        when_not_equal: Literal["nonmatch", "indefinite"],
    ) -> None:
        """Create an IDLinker

        Parameters
        ----------
        labels:
            Something that resolves against the left and right table to the Column
            that holds the labels.
        when_not_equal:
            What to do for record pairs where the labels are not equal.
            If "nonmatch", then we will consider the records to be a non-match.
            If "indefinite", then we don't know if the records are a match or not,
            and will leave that decision to other parts of the pipeline.
        when_null:
            What to do for record pairs where  one or both of them have a null label.
            If "nonmatch", then we will consider the records to be a non-match.
            If "indefinite", then we don't know if the records are a match or not,
            and will leave that decision to other parts of the pipeline.
        """
        if when_null not in ("nonmatch", "indefinite"):
            raise ValueError(
                f"when_null should be either 'nonmatch' or 'indefinite'. Got {when_null}"  # noqa: E501
            )
        if when_not_equal not in ("nonmatch", "indefinite"):
            raise ValueError(
                f"when_not_equal should be either 'nonmatch' or 'indefinite'. Got {when_not_equal}"  # noqa: E501
            )

        self.resolvers = tuple(_resolve.key_pair_resolvers(labels))
        self.when_null = when_null
        self.when_not_equal = when_not_equal

    def match_condition(self, a: ibis.Table, b: ibis.Table) -> ir.BooleanColumn:
        """Select any pairs where we know they are a match (ie the labels are equal)."""
        return KeyLinker(self.resolvers).__join_condition__(a, b)

    def match_linkage(self, left: ibis.Table, right: ibis.Table) -> Linkage:
        if right is left:
            right = right.view()
        links = LinksTable.from_join_condition(left, right, self.match_condition)
        return Linkage(left=left, right=right, links=links)

    # def nonmatch_condition(self, a: ibis.Table, b: ibis.Table) -> ir.BooleanValue:
    #     """Select any pairs where we know they are a non-match.

    #     Includes pairs where either
    #     - one or both of the labels are null (iff `when_null` is "nonmatch")
    #     - the labels are not equal (iff `when_not_equal` is "nonmatch")
    #     """
    #     label_a, label_b = _resolve.resolve_column_pair(self.labels, a, b)
    #     conditions = []
    #     if self.when_not_equal == "nonmatch":
    #         conditions.append(
    #             ibis.and_(
    #                 label_a != label_b,
    #                 label_a.notnull(),
    #                 label_b.notnull(),
    #             )
    #         )
    #     if self.when_null == "nonmatch":
    #         conditions.append(label_a.isnull() | label_b.isnull())
    #     if not conditions:
    #         return ibis.literal(False)
    #     return ibis.or_(*conditions)

    def indefinite_condition(self, a: ibis.Table, b: ibis.Table) -> ir.BooleanValue:
        """Select any pairs where they are not a match and they are not a non-match."""
        raise NotImplementedError()

    def indefinite_linkage(self, left: ibis.Table, right: ibis.Table) -> Linkage:
        if right is left:
            right = right.view()
        links = LinksTable.from_join_condition(left, right, self.indefinite_condition)
        return Linkage(left=left, right=right, links=links)

    def __repr__(self) -> str:
        return f"IDLinker(labels={self.resolvers}, when_null={self.when_null}, when_not_equal={self.when_not_equal})"  # noqa: E501



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from ._weights import ComparerWeights

if TYPE_CHECKING:
    import altair as alt


def log_odds_color_scale() -> alt.Scale:
    return alt.Scale(
        domainMid=0,
        domainMin=-3,
        domainMax=3,
        scheme="redyellowgreen",
    )


def plot_weights(weights: ComparerWeights | Iterable[ComparerWeights]) -> alt.Chart:
    """Plot the weights for Comparer(s).

    Use this to
    - See which levels are common and which are rare.
      If all pairs are getting matched by only one level, you probably
      want to adjust the conditions so that pairs are more evenly distributed.
      For example, if you have an "exact match" level that hardly is ever used,
      that could be an indication that your condition is too strict and you should
      relax it.
    - See the odds for each level.
      If the odds for a "exact match" level are lower than you expect, perhaps
      near 1, that could be an indication that your condition is too loose and
      there are many non-matches sneaking into that level. You should inspect
      those pairs and figure out how to tighted the condition so
      that only matches are in that level.

    Parameters
    ----------
    weights :
        The weights to plot.

    Returns
    -------
    alt.Chart
        The plot.
    """
    if isinstance(weights, ComparerWeights):
        weights = [weights]
    subplots = [_plot_comparer_weights(cw) for cw in weights]
    together = alt.vconcat(*subplots, spacing=30)
    return together.interactive()


def _plot_comparer_weights(cw: ComparerWeights) -> alt.Chart:
    import altair as alt
    import pandas as pd

    records = []
    for i, lw in enumerate(cw):
        records.append(
            {
                "level": lw.name,
                "m": lw.m,
                "u": lw.u,
                "odds": lw.odds,
                "log_odds": lw.log_odds,
                "level_order": i,
            }
        )
    df = pd.DataFrame(records)

    mu_width = 200
    ms = _subplot(
        df,
        alt.X(
            "m",
            title=["Proportion of Pairs", "Amongst Matches"],
            axis=alt.Axis(format="%"),
            scale=alt.Scale(domain=[0, 1]),
        ),
        alt.value("green"),
        True,
        mu_width,
    )
    us = _subplot(
        df,
        alt.X(
            "u",
            title=["Proportion of Pairs", "Amongst Non-Matches"],
            axis=alt.Axis(format="%"),
            scale=alt.Scale(domain=[0, 1]),
        ),
        alt.value("red"),
        False,
        mu_width,
    )
    odds = _subplot(
        df,
        alt.X(
            "log_odds",
            title="Odds",
            scale=alt.Scale(domain=[-3, 3]),
            axis=alt.Axis(labelExpr=10**alt.datum.value),
        ),
        alt.Color(
            "log_odds",
            title="Log Odds",
            scale=log_odds_color_scale(),
            legend=None,
        ),
        False,
        width=250,
    )
    together = alt.hconcat(ms, us, odds, spacing=5)
    together = together.properties(
        title=alt.Title(text=f"Weights for Comparer '{cw.name}'", anchor="middle")
    )
    return together


def _subplot(t, x, color, use_y_axis, width):
    import altair as alt

    axis = alt.Axis() if use_y_axis else None
    y = alt.Y("level", title=None, sort=alt.EncodingSortField("level_order"), axis=axis)
    t = t.copy()
    t["explanation"] = _odds_explanation(t["odds"])
    chart = (
        alt.Chart(t, width=width)
        .encode(
            x=x,
            y=y,
            tooltip=[
                alt.Tooltip("level", title="Level"),
                alt.Tooltip("m", title="Proportion Amongst Matching Pairs", format="%"),
                alt.Tooltip(
                    "u", title="Proportion Amongst Non-Matching Pairs", format="%"
                ),
                alt.Tooltip("odds", title="Odds", format="f"),
                alt.Tooltip("explanation", title="Explanation"),
            ],
            color=color,
        )
        .mark_bar()
    )
    return chart


def _odds_explanation(odds):
    def format_odds(odds):
        return odds.apply(lambda float_: f"{float_:,.3f}")

    more = (
        "Seeing this level makes the odds of a match "
        + format_odds(odds)
        + " times more likely"
    )
    less = (
        "Seeing this level makes the odds of a match "
        + format_odds(1 / odds)
        + " times less likely"
    )
    return more.where((odds >= 1), less)



from __future__ import annotations

import math
from typing import Union, overload

import ibis
from ibis import Deferred
from ibis.expr import types as ir

_IbisIsh = Union[Deferred, ir.NumericValue]


@overload
def prob_to_odds(prob: float) -> float: ...


@overload
def prob_to_odds(prob: _IbisIsh) -> _IbisIsh: ...


def prob_to_odds(prob):
    normal = prob / (1 - prob)
    if isinstance(prob, (Deferred, ir.NumericValue)):
        return ibis.ifelse(prob != 1, normal, ibis.literal(float("inf")))
    else:
        return normal if prob != 1 else float("inf")


@overload
def odds_to_prob(odds: float) -> float: ...


@overload
def odds_to_prob(odds: _IbisIsh) -> _IbisIsh: ...


def odds_to_prob(odds):
    normal = odds / (1 + odds)
    if isinstance(odds, (Deferred, ir.NumericValue)):
        return ibis.ifelse(odds.isinf(), 1.0, normal)
    else:
        return 1 if odds == float("inf") else normal


@overload
def odds_to_log_odds(odds: float) -> float: ...


@overload
def odds_to_log_odds(odds: _IbisIsh) -> _IbisIsh: ...


def odds_to_log_odds(odds):
    if isinstance(odds, (Deferred, ir.NumericValue)):
        return odds.log10()
    else:
        if odds == 0:
            return float("-inf")
        else:
            return math.log10(odds)



from __future__ import annotations

from mismo.types._diff import Diff as Diff
from mismo.types._diff import DiffStats as DiffStats
from mismo.types._linked_table import LinkCountsTable as LinkCountsTable
from mismo.types._linked_table import LinkedTable as LinkedTable
from mismo.types._links_table import LinksTable as LinksTable
from mismo.types._union_table import UnionTable as UnionTable
from mismo.types._updates import Updates as Updates



"""Extra utils for working with ibis arrays."""

from __future__ import annotations

from mismo.arrays._array import array_choice as array_choice
from mismo.arrays._array import array_combinations as array_combinations
from mismo.arrays._array import (
    array_combinations_first_n as array_combinations_first_n,
)
from mismo.arrays._array import array_filter_isin_other as array_filter_isin_other
from mismo.arrays._array import array_shuffle as array_shuffle
from mismo.arrays._array import array_sort as array_sort
from mismo.arrays._builtins import array_all as array_all
from mismo.arrays._builtins import array_any as array_any
from mismo.arrays._builtins import array_max as array_max
from mismo.arrays._builtins import array_mean as array_mean
from mismo.arrays._builtins import array_median as array_median
from mismo.arrays._builtins import array_min as array_min
from mismo.arrays._builtins import array_sum as array_sum



from __future__ import annotations

import ibis
from ibis.expr import types as ir

from mismo import _util


def double_metaphone(s: ir.StringValue) -> ir.ArrayValue[ir.StringValue]:
    """Double Metaphone phonetic encoding

    This requires the [doublemetaphone](https://github.com/dedupeio/doublemetaphone)
    package to be installed.
    You can install it with `python -m pip install DoubleMetaphone`.
    This uses a python UDF so it is going to be slow.

    Examples
    --------
    >>> from mismo.text import double_metaphone
    >>> double_metaphone("catherine").execute()
    ['K0RN', 'KTRN']
    >>> double_metaphone("").execute()
    ['', '']
    >>> double_metaphone(None).execute() is None
    True
    """
    s = _util.ensure_ibis(s, "string")
    return _dm_udf(s)


@ibis.udf.scalar.python(signature=(("string",), "array<string>"))
def _dm_udf(s):
    with _util.optional_import("DoubleMetaphone"):
        from doublemetaphone import doublemetaphone

    return doublemetaphone(s)


# TODO: this isn't portable between backends
@ibis.udf.scalar.builtin
def damerau_levenshtein(a: str, b: str) -> int:
    """
    The number of adds, deletes, substitutions, and transposes to get from `a` to `b`.

    This is the levenstein distance with the addition of transpositions as
    a possible operation.
    """


def levenshtein_ratio(s1: ir.StringValue, s2: ir.StringValue) -> ir.FloatingValue:
    """The levenshtein distance between two strings, normalized to be between 0 and 1.

    The ratio is defined as `(lenmax - ldist)/lenmax` where

    - `ldist` is the regular levenshtein distance
    - `lenmax` is the maximum length of the two strings
      (eg the largest possible edit distance)

    This makes it so that the ratio is 1 when the strings are the same and 0
    when they are completely different.
    By doing this normalization, the ratio is always between 0 and 1, regardless
    of the length of the strings.

    Parameters
    ----------
    s1:
        The first string

    s2:
        The second string

    Returns
    -------
    lev_ratio:
        The ratio of the Levenshtein edit cost to the maximum string length

    Examples
    --------
    >>> from mismo.text import levenshtein_ratio
    >>> levenshtein_ratio("mile", "mike").execute()
    0.75
    >>> levenshtein_ratio("mile", "mile").execute()
    1.0
    >>> levenshtein_ratio("mile", "").execute()
    0.0
    >>> levenshtein_ratio("", "").execute()
    nan
    """
    return _dist_ratio(s1, s2, lambda a, b: a.levenshtein(b))


def damerau_levenshtein_ratio(
    s1: ir.StringValue, s2: ir.StringValue
) -> ir.FloatingValue:
    """Like levenshtein_ratio, but with the Damerau-Levenshtein distance.

    See Also
    --------
    - [damerau_levenshtein()][mismo.text.damerau_levenshtein]
    - [levenshtein_ratio()][mismo.text.levenshtein_ratio]
    """
    return _dist_ratio(s1, s2, damerau_levenshtein)


def _dist_ratio(s1, s2, dist):
    s1 = _util.ensure_ibis(s1, "string")
    s2 = _util.ensure_ibis(s2, "string")
    lenmax = ibis.greatest(s1.length(), s2.length())
    return (lenmax - dist(s1, s2)) / lenmax


@ibis.udf.scalar.builtin(name="jaro_similarity")
def _jaro_similarity(s1: str, s2: str) -> float: ...


def jaro_similarity(s1: ir.StringValue, s2: ir.StringValue) -> ir.FloatingValue:
    """The jaro similarity between `s1` and `s2`.

    This is a number between 0 and 1, defined as
    `sj = 1/3 * (m/l_1 + m/l_2 + (m-t)/m)`

    where `m` is the number of matching characters between s1 and s2 and `t` is the
    number of transpositions between `s1` and `s2`.

    Examples
    --------
    >>> import ibis
    >>> from mismo.text import jaro_similarity
    >>> jaro_similarity(ibis.literal("foo"), ibis.literal("foo")).execute()
    1.0
    >>> jaro_similarity(ibis.literal("foo"), ibis.literal("food")).execute()
    0.9166666666666666
    >>> jaro_similarity(ibis.null(str), ibis.literal("food")).execute()
    nan

    Be aware: comparing to an empty string always has a similarity of 0:

    >>> jaro_similarity(ibis.literal("a"), ibis.literal("")).execute()
    0.0
    >>> jaro_similarity(ibis.literal(""), ibis.literal("")).execute()
    0.0
    """
    return _jaro_similarity(s1, s2)


# TODO: This isn't portable between backends
@ibis.udf.scalar.builtin(name="jaro_winkler_similarity")
def _jaro_winkler_similarity(s1: str, s2: str) -> float: ...


def jaro_winkler_similarity(s1: ir.StringValue, s2: ir.StringValue) -> ir.FloatingValue:
    """The Jaro-Winkler similarity between `s1` and `s2`.

    The Jaro-Winkler similarity is a variant of the Jaro similarity that
    measures the number of edits between two strings
    and places a higher importance on the prefix.

    It is defined as `(sjw = sj + l * p * (1-sj)`
    where `sj` is the Jaro similarity, `l` is the length of the common prefix  (up to a
    maximum of 4) and `p` is a constant scaling factor (up to a maximum of 0.25, but
    typically set to 0.1)

    Examples
    --------
    >>> import ibis
    >>> from mismo.text import jaro_winkler_similarity
    >>> jaro_winkler_similarity(ibis.literal("foo"), ibis.literal("foo")).execute()
    1.0
    >>> jaro_winkler_similarity(ibis.literal("foo"), ibis.literal("food")).execute()
    0.9416666666666667
    >>> jaro_winkler_similarity(ibis.null(str), ibis.literal("food")).execute()
    nan

    Be aware: comparing to an empty string always has a similarity of 0:

    >>> jaro_winkler_similarity(ibis.literal("a"), ibis.literal("")).execute()
    0.0
    >>> jaro_winkler_similarity(ibis.literal(""), ibis.literal("")).execute()
    0.0
    """
    return _jaro_winkler_similarity(s1, s2)



from __future__ import annotations

from typing import Callable, Iterable, Mapping, Sequence

import ibis
from ibis.expr import types as ir


class Datasets:
    """An ordered, dict-like collection of tables of records.

    All tables must have a column named 'record_id' that is globally unique.
    The dtype of the 'record_id' column must be the same in all tables.
    Besides that, the schema of the tables can be different.

    This is a nice abstraction over the fact that some record linkage problems
    are deduplication, and thus involve only one table, while others are
    linkage, and involve two tables.
    """

    def __init__(
        self, tables: ir.Table | Iterable[ir.Table] | Mapping[str, ir.Table]
    ) -> None:
        """Create a new Datasets.

        If `tables` is Mapping, then it is used as-is.
        If `tables` is a single table, it is named "dataset_0".
        If `tables` is an iterable of tables, then we try to find their names by:
        - calling `get_name()` on each table. If that fails, then fall back to...
        - using "left" and "right" for two tables
        - using "dataset_i" otherwise
        """
        if isinstance(tables, ir.Table):
            tables = {"dataset_0": tables}
        tables = _get_names(tables)

        self._tables = {}
        for name, t in tables.items():
            if "record_id" not in t.columns:
                raise ValueError("records must have a record_id column")
            # should we .cache() here to guarantee consistency?
            self._tables[name] = t
        if "record_id" not in self.shared_schema:
            raise ValueError("The record_id column must be the same type in all tables")

    def __len__(self) -> int:
        """The number of tables."""
        return len(self._tables)

    def __getitem__(self, key: str | int) -> ir.Table:
        """Get a table by name or index."""
        if isinstance(key, int):
            return tuple(self)[key]
        return self._tables[key]

    def __iter__(self) -> Iterable[ir.Table]:
        """Iterate over the tables in the order they were added."""
        return iter(self._tables.values())

    def __contains__(self, key: str | ir.Table) -> bool:
        """Check if a table is in the collection by name or value."""
        if isinstance(key, str):
            return key in self._tables
        return key in self._tables.values()

    def __repr__(self) -> str:
        return repr({name: table.head(5) for name, table in self.items()})

    def cache(self) -> Datasets:
        """Return a new Datasets with all tables cached."""
        return self.__class__({name: t.cache() for name, t in self.items()})

    def map(self, f: ibis.Deferred | Callable[[str, ir.Table], ir.Table]) -> ir.Table:
        """Return a new Datasets with all tables transformed by `f`."""
        if isinstance(f, ibis.Deferred):
            return self.__class__({name: f.bind(t) for name, t in self.items()})
        else:
            return self.__class__({name: f(name, t) for name, t in self.items()})

    def filter(
        self, f: ibis.Deferred | Callable[[str, ir.Table], ir.Table]
    ) -> ir.Table:
        """Return a new Datasets with all tables filtered by `f`."""
        if isinstance(f, ibis.Deferred):
            return self.__class__({name: t.filter(f) for name, t in self.items()})
        else:
            return self.__class__(
                {name: t.filter(f(name, t)) for name, t in self.items()}
            )

    @property
    def names(self) -> tuple[str, ...]:
        """The names of the underlying tables."""
        return tuple(self._tables.keys())

    @property
    def tables(self) -> tuple[ir.Table, ...]:
        """The underlying tables."""
        return tuple(self)

    def keys(self) -> Iterable[str]:
        """The names of the underlying tables."""
        return self._tables.keys()

    def values(self) -> Iterable[ir.Table]:
        """The underlying tables."""
        return self._tables.values()

    def items(self) -> Iterable[tuple[str, ir.Table]]:
        """The names and tables of the underlying tables."""
        return self._tables.items()

    @property
    def shared_schema(self) -> ibis.Schema:
        """The schema that all tables have in common.

        Columns with conflicting types are omitted.

        This is useful for operations that require the same schema in all tables,
        for example getting all the record_ids.
        """
        dtypes_per_col = {}
        for t in self:
            for col, dtype in t.schema().items():
                if col not in dtypes_per_col:
                    dtypes_per_col[col] = [dtype]
                else:
                    dtypes_per_col[col].append(dtype)
        good = {
            col: dtypes[0]
            for col, dtypes in dtypes_per_col.items()
            if len(dtypes) == len(self) and len(set(dtypes)) == 1
        }
        return ibis.Schema(good)

    def unioned(self) -> ir.Table:
        """Select the `self.shared_schema` columns from all tables and union them."""
        return ibis.union(
            *(
                t.select(*self.shared_schema).mutate(dataset=ibis.literal(name))
                for name, t in self.items()
            )
        )

    def all_record_ids(self) -> ir.Column:
        """Return all unique record_ids from all tables."""
        return self.unioned().select("record_id").distinct().record_id


def _get_names(
    tables: Sequence[ir.Table] | Mapping[str, ir.Table],
) -> dict[str, ir.Table]:
    def _get_name(i: int, t: ir.Table) -> str:
        try:
            ds = t.get_name()
        except AttributeError:
            pass
        else:
            if not ds.startswith("ibis_cache"):
                return ds

        if len(tables) == 2:
            return "left" if i == 0 else "right"
        else:
            return f"dataset_{i}"

    try:
        return dict(tables.items())
    except AttributeError:
        return {_get_name(i, t): t for i, t in enumerate(tables)}



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Callable, Protocol, TypeAlias, Union, runtime_checkable

import ibis
from ibis import Deferred
from ibis.common.deferred import var
from ibis.expr import types as ir

from mismo import _registry, _resolve, _util


@runtime_checkable
class HasJoinCondition(Protocol):
    """
    Has `__join_condition__(left: ibis.Table, right: ibis.Table)`, which returns something that `ibis.join()` understands.

    There are concrete implementations of this for various types of join conditions.
    For example, `BooleanJoinCondition` wraps a boolean
    or an `ibis.ir.BooleanValue` expression.
    """  # noqa: E501

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue | bool:
        """
        Given a left and right table, return something that `ibis.join()` understands.
        """
        pass


IntoHasJoinCondition: TypeAlias = Union[
    HasJoinCondition,
    bool,
    ibis.ir.BooleanValue,
    str,
    ibis.Deferred,
    Callable[[ibis.Table, ibis.Table], HasJoinCondition | bool | ibis.ir.BooleanValue],
    Iterable[
        HasJoinCondition,
        bool,
        ibis.ir.BooleanValue,
        str,
        ibis.Deferred,
        Callable[
            [ibis.Table, ibis.Table], HasJoinCondition | bool | ibis.ir.BooleanValue
        ],
    ],
]
"""
An object that can be converted into a HasJoinCondition with mismo.join_condition()
"""


class JoinConditionRegistry(
    _registry.Registry[Callable[..., HasJoinCondition], HasJoinCondition]
):
    """A registry for HasJoinCondition factory functions"""


_join_condition_registry = JoinConditionRegistry()


@_join_condition_registry.register
def _already_has_join_condition(obj: Any) -> HasJoinCondition:
    """If the object already has a `__join_condition__` method, return it as is."""
    if isinstance(obj, HasJoinCondition):
        return obj
    return NotImplemented


def join_condition(obj: IntoHasJoinCondition) -> HasJoinCondition:
    """
    Create a [HasJoinCondition][mismo.HasJoinCondition] from an object.

    Parameters
    ----------
    obj
        The object to create a join condition from.
        This can be anything that ibis understands as join condition,
        such as a boolean, an ibis.ir.BooleanValue expression, a `str`,
        an ibis.Deferred, etc.
        It also supports other types,
        such as `lambda left, right: <one of the above>`,

    Returns
    -------
        An object that follows the [HasJoinCondition][mismo.HasJoinCondition] protocol.
    """
    return _join_condition_registry(obj)


join_condition.register = _join_condition_registry.register


class BooleanJoinCondition:
    def __init__(self, boolean: bool | ir.BooleanValue):
        self.boolean = boolean

    @join_condition.register
    @staticmethod
    def _try(obj: Any) -> BooleanJoinCondition:
        if not isinstance(obj, (bool, ir.BooleanValue)):
            return NotImplemented
        return BooleanJoinCondition(obj)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue | bool:
        return self.boolean

    def __repr__(self):
        return f"{self.__class__.__name__}({self.boolean!r})"


class FuncJoinCondition:
    def __init__(self, func: Callable[[ibis.Table, ibis.Table], ibis.ir.BooleanValue]):
        self.func = func

    @join_condition.register
    @staticmethod
    def _try(obj: Any) -> FuncJoinCondition:
        # Deffered's think they are callable, so guard against that.
        if isinstance(obj, ibis.Deferred):
            return NotImplemented
        if not callable(obj):
            return NotImplemented
        return FuncJoinCondition(obj)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue:
        return self.func(left, right)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.func!r})"


class AndJoinCondition:
    def __init__(self, subconditions: Iterable[Any]):
        self.subconditions: tuple[HasJoinCondition] = tuple(
            join_condition(c) for c in subconditions
        )

    @join_condition.register
    @staticmethod
    def _try(obj: Any) -> AndJoinCondition:
        tuple_subconditions = _try_iterable(obj)
        if tuple_subconditions is None:
            return NotImplemented
        return AndJoinCondition(tuple_subconditions)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue:
        return self.join_condition(left, right)

    def join_condition(self, left: ibis.Table, right: ibis.Table) -> ir.BooleanColumn:
        conditions = [c.__join_condition__(left, right) for c in self.subconditions]
        return ibis.and_(*conditions)


# KeyJoinCondition needs to be registered afterwards so that it has
# priority with 2-tuples
class MultiKeyJoinCondition:
    """Join where all of the keys match."""

    def __init__(self, subconditions: Iterable[Any]):
        self.subconditions: tuple[KeyJoinCondition] = tuple(
            join_condition(c) for c in subconditions
        )

    @join_condition.register
    def _try(obj: Any) -> MultiKeyJoinCondition:
        subconditions = _try_iterable(obj)
        if subconditions is None:
            return NotImplemented
        resolved = [join_condition(c) for c in subconditions]
        for sub in resolved:
            if not isinstance(sub, KeyJoinCondition):
                return NotImplemented
        return MultiKeyJoinCondition(resolved)

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue:
        return self.join_condition(left, right)

    def join_condition(self, left: ibis.Table, right: ibis.Table) -> ir.BooleanColumn:
        conditions = [c.__join_condition__(left, right) for c in self.subconditions]
        return ibis.and_(*conditions)


class BadKeyJoinCondition(ValueError):
    pass


class KeyJoinCondition:
    def __init__(self, spec: str | Deferred | tuple[str | Deferred, str | Deferred]):
        if isinstance(spec, str):
            left_spec = right_spec = spec
        elif isinstance(spec, Deferred):
            if _resolve.variables_names(spec) != {"_"}:
                raise BadKeyJoinCondition(
                    "Deferred join conditions must only contain `ibis._` variables."
                )
            left_spec = right_spec = spec
        elif isinstance(spec, tuple):
            if len(spec) != 2:
                # This is eg ("name", "age", "address"), raise this so that
                # AndJoinCondition picks it up.
                raise BadKeyJoinCondition(spec)
            left_spec, right_spec = spec
        else:
            raise BadKeyJoinCondition(spec)

        if not isinstance(left_spec, (str, Deferred)):
            raise BadKeyJoinCondition(spec)
        if not isinstance(right_spec, (str, Deferred)):
            raise BadKeyJoinCondition(spec)
        self.left_spec = left_spec
        self.right_spec = right_spec

    @join_condition.register
    @staticmethod
    def _try(obj: Any) -> KeyJoinCondition:
        try:
            return KeyJoinCondition(obj)
        except BadKeyJoinCondition:
            return NotImplemented

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ir.BooleanValue:
        return self.join_condition(left, right)

    def bind_columns(
        self, left: ibis.Table, right: ibis.Table
    ) -> list[tuple[ibis.Column, ibis.Column]]:
        # print(self.left_spec, self.right_spec)
        l_cols = _util.bind(left, self.left_spec)
        r_cols = _util.bind(right, self.right_spec)
        # print(self.left_spec, self.right_spec)
        return list(zip(l_cols, r_cols, strict=True))

    def join_condition(self, left: ibis.Table, right: ibis.Table) -> ir.BooleanColumn:
        conditions = [coll == colr for coll, colr in self.bind_columns(left, right)]
        return ibis.and_(*conditions)


left = var("left")
"""A deferred placeholder for the left table in a join.

Examples
--------
>>> condition = mismo.left.last_name.upper() == mismo.right.family_name.upper()
>>> import ibis
>>> my_left_table = ibis.memtable([("johnson",), ("smith",)], columns=["last_name"])
>>> my_right_table = ibis.memtable([("JOHNSON",), ("JONES",)], columns=["family_name"])
>>> mismo.join(my_left_table, my_right_table, condition)
┏━━━━━━━━━━━┳━━━━━━━━━━━━━┓
┃ last_name ┃ family_name ┃
┡━━━━━━━━━━━╇━━━━━━━━━━━━━┩
│ string    │ string      │
├───────────┼─────────────┤
│ johnson   │ JOHNSON     │
└───────────┴─────────────┘
"""
right = var("right")
"""A deferred placeholder for the right table in a join.

Examples
--------
>>> condition = mismo.left.last_name.upper() == mismo.right.family_name.upper()
>>> import ibis
>>> my_left_table = ibis.memtable([("johnson",), ("smith",)], columns=["last_name"])
>>> my_right_table = ibis.memtable([("JOHNSON",), ("JONES",)], columns=["family_name"])
>>> mismo.join(my_left_table, my_right_table, condition)
┏━━━━━━━━━━━┳━━━━━━━━━━━━━┓
┃ last_name ┃ family_name ┃
┡━━━━━━━━━━━╇━━━━━━━━━━━━━┩
│ string    │ string      │
├───────────┼─────────────┤
│ johnson   │ JOHNSON     │
└───────────┴─────────────┘
"""


class LeftRightDeferredCondition:
    def __init__(self, condition: Deferred):
        if _resolve.variables_names(condition) != {"left", "right"}:
            raise ValueError(
                f"{self.__class__.__name__} must only contain 'left' and 'right'"
                "variables, eg `mismo.left.last_name == mismo.right.family_name`."
            )
        self.condition = condition

    @join_condition.register
    @staticmethod
    def _try(obj: Any) -> LeftRightDeferredCondition:
        if not isinstance(obj, Deferred):
            return NotImplemented
        try:
            return LeftRightDeferredCondition(obj)
        except ValueError:
            return NotImplemented

    def __join_condition__(
        self, left: ibis.Table, right: ibis.Table
    ) -> ibis.ir.BooleanValue:
        return self.condition.resolve(left=left, right=right)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.condition!r})"


def _try_iterable(obj: Any) -> tuple | None:
    if isinstance(obj, (str, bytes, bytearray, Deferred)):
        return None
    try:
        return tuple(c for c in obj)
    except TypeError as e:
        if "is not iterable" in str(e):
            return None
        raise



from __future__ import annotations

from ibis.expr import types as ir


def jaccard(a: ir.ArrayValue, b: ir.ArrayValue) -> ir.FloatingValue:
    """The Jaccard similarity between two arrays.

    Parameters
    ----------
    a : ArrayValue
        The first array.
    b : ArrayValue
        The second array.

    Returns
    -------
    FloatingValue
        The Jaccard similarity between the two arrays.
    """
    intersection = a.intersect(b).length()
    normal = intersection / a.union(b).length()
    return (intersection == 0).ifelse(0, normal)



from __future__ import annotations

from ibis.expr import types as ir


def mutate(struct: ir.StructValue, **kwargs: ir.Value) -> ir.StructValue:
    """Mutate a struct by adding or replacing columns.

    Analogous to ibis.Table.mutate(**kwargs).
    """
    default = {field: struct[field] for field in struct.fields}
    return ir.struct({**default, **kwargs})


def drop(struct: ir.StructValue, *fields: str) -> ir.StructValue:
    """Mutate a struct by dropping columns.

    Analogous to ibis.Table.drop(fields)."""
    new_fields = {
        field: struct[field] for field in struct.fields if field not in fields
    }
    return ir.struct(new_fields)


def select(struct: ir.Column, *fields: str) -> ir.Column:
    """Select columns from a struct.

    Analogous to ibis.Table.select(fields)."""
    return ir.struct({field: struct[field] for field in fields})


def rename(struct: ir.StructValue, **renamings: str) -> ir.StructValue:
    """Rename the fields in a struct, analogous to Table.rename()"""
    fields = {field: struct[field] for field in struct.fields}
    for new, old in renamings.items():
        fields[new] = fields.pop(old)
    return ir.struct(fields)


def unpack(struct: ir.StructValue) -> tuple[ir.Value, ...]:
    """Unpack the values of a struct into a tuple.

    Replacement for the deprecated struct.destructure() method,
    and analogous to the Table.unpack("my_struct_col") method.
    """
    return (struct[field_name].name(field_name) for field_name in struct.type().names)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, Literal

import ibis
from ibis.expr import datatypes as dt
from ibis.expr import types as ir

from mismo import _util, joins
from mismo.types._table_wrapper import TableWrapper

if TYPE_CHECKING:
    from collections.abc import Mapping


class Filters:
    """A simple namespace for filter functions."""

    @staticmethod
    def all_different(subset: Iterable[str] | None = None):
        """
        Make a Updates filter function that gives rows where all columns are different.

        Parameters
        ----------
        subset : Iterable[str] | None
            The columns to consider.

            If None, and the schemas of the before and after tables are dufferent,
            then we return True (ie, keep all rows), because by definition
            every row is different, since the schemas are different.

            If None, and the schemas of the before and after tables are the same,
            then we consider all columns in both before and after tables.

            If you joined on an eg id column, you almost definitely want to exclude it here.

        Examples
        --------
        >>> u = Updates.from_tables(before, after, join_on="id")  # doctest: +SKIP
        >>> u.filter(u.filters.all_different(["name", "age"]))  # doctest: +SKIP
        >>> u.filter(
        ...     u.filters.all_different([c for c in u.columns if c != "my_id"])
        ... )  # doctest: +SKIP
        """  # noqa: E501

        def filter_func(table: ir.Table):
            nonlocal subset
            if subset is None:
                u = Updates(table, check_schemas="lax")
                if u.before().schema() != u.after().schema():
                    return True
                subset = table.columns
            return ibis.and_(*(is_changed(table[col]) for col in subset))

        return filter_func

    @staticmethod
    def any_different(subset: Iterable[str] | None = None):
        """Make a Updates filter function that gives rows where any column is different.

        Parameters
        ----------
        subset : Iterable[str] | None
            The columns to consider.

            If None, and the schemas of the before and after tables are dufferent,
            then we return True (ie, keep all rows), because by definition
            every row is different, since the schemas are different.

            If None, and the schemas of the before and after tables are the same,
            then we consider all columns in both before and after tables.

        Examples
        --------
        >>> u = Updates.from_tables(before, after, join_on="id")  # doctest: +SKIP
        >>> u.filter(u.filters.any_different(["name", "age"]))  # doctest: +SKIP
        """

        def filter_func(table: ir.Table):
            nonlocal subset
            if subset is None:
                u = Updates(table, check_schemas="lax")
                if u.before().schema() != u.after().schema():
                    return True
                subset = table.columns
            return ibis.or_(*(is_changed(table[col]) for col in subset))

        return filter_func


class Updates(TableWrapper):
    """A Table representing how individual rows were updated.

    This only represents differences in rows that exist both before and after the changes.
    To represent a general difference between two tables, eg with insertions and deletions,
    use [Diff](mismo.Diff).

    This represents how each column has changed between two tables.
    If a column only has a 'before' field, it means this column was removed.
    If a column only has an 'after' field, it means this column was added.
    If a column has both 'before' and 'after' fields, it means this column was present in both tables.
    """  # noqa: E501

    filters = Filters
    """A set of filters for convenience.
    
    Examples
    --------
    >>> u = Updates.from_tables(before, after, join_on="id")  # doctest: +SKIP
    >>> u.filter(u.filters.all_different(["name", "age"]))  # doctest: +SKIP
    """

    def __init__(
        self,
        diff_table: ibis.Table,
        /,
        *,
        check_schemas: Literal["exactly", "names", "lax"] = "exactly",
    ) -> None:
        """Create an Updates object from a table of differences.

        Parameters
        ----------
        diff_table : ibis.Table
            A table with columns that are structs with at least 'before' and 'after' fields.
        check_schemas : Literal["exactly", "names", "lax"]
            How to check the schemas of the before and after values.
            - "exactly": both before and after must have the same columns and types.
            - "names": both before and after must have the same columns, but types can differ.
            - "lax": no schema checking, just that there is at least one of 'before' or 'after' in each column.
        """  # noqa: E501

        def _check_col(name: str):
            col_type: dt.DataType = diff_table[name].type()
            if not isinstance(col_type, dt.Struct):
                return f"Column {name} is not a struct"
            fields = set(col_type.names)
            if "before" not in fields and "after" not in fields:
                return f"Column {name} needs to have at least one of 'before' and 'after' field"  # noqa: E501
            if check_schemas == "exactly":
                if "before" not in fields or "after" not in fields:
                    return (
                        f"Column {name} needs to have both 'before' and 'after' fields"
                    )
                b_type = col_type.fields["before"]
                a_type = col_type.fields["after"]
                if b_type != a_type:
                    return f"Column {name} needs to have the same type for 'before' ({b_type}) and 'after' ({a_type})"  # noqa: E501
            elif check_schemas == "names":
                if "before" not in fields or "after" not in fields:
                    return (
                        f"Column {name} needs to have both 'before' and 'after' fields"
                    )
            elif check_schemas == "lax":
                pass
            else:
                raise ValueError(f"Unknown schema {check_schemas}")

        errors = [_check_col(col_name) for col_name in diff_table.columns]
        errors = [e for e in errors if e is not None]
        if errors:
            raise ValueError("\n".join(errors))

        object.__setattr__(self, "_check_schemas", check_schemas)
        super().__init__(diff_table)

    @property
    def check_schemas(self) -> Literal["exactly", "names", "lax"]:
        """The schema checking mode used for this Updates object."""
        return self._check_schemas

    @classmethod
    def from_tables(
        cls,
        before: ibis.Table,
        after: ibis.Table,
        *,
        join_on: str | Literal[False],
        check_schemas: Literal["exactly", "names", "lax"] = "exactly",
    ) -> Updates:
        """Create from two different tables by joining them on a key.

        Note that this results in only the rows that are present in both tables,
        due to the inner join on the key. Insertions and deletions should be
        handled separately.
        """
        # Prefer a column order of
        # 1. all the columns in after
        # 2. any extra columns in before are tacked on the end
        if after is before:
            after = after.view()
        all_columns = (dict(before.schema()) | dict(after.schema())).keys()
        joined = joins.join(
            before, after, join_on, lname="{name}_l", rname="{name}_r", rename_all=True
        )

        def make_diff_col(col: str) -> ir.StructColumn:
            d = {}
            col_l = col + "_l"
            col_r = col + "_r"
            if col_l in joined.columns:
                # need to do this cast because nonnull dtypes
                # become nullable after the join.
                d["before"] = joined[col_l].cast(before.schema()[col])
            if col_r in joined.columns:
                d["after"] = joined[col_r].cast(after.schema()[col])
            assert d, f"Column {col} not found in either before or after"
            return ibis.struct(d).name(col)

        diff_table = joined.select(*[make_diff_col(c) for c in all_columns])
        return cls(diff_table, check_schemas=check_schemas)

    @classmethod
    def from_before_after(
        cls,
        before: Mapping[str, ibis.Value] | ir.StructValue,
        after: Mapping[str, ibis.Value] | ir.StructValue,
        *,
        check_schemas: Literal["exactly", "names", "lax"] = "exactly",
    ) -> Updates:
        """Create an Updates object from before and after values.

        Parameters
        ----------
        before : Mapping[str, ibis.Value] | ir.StructValue
            The values before the changes.
        after : Mapping[str, ibis.Value] | ir.StructValue
            The values after the changes.
        check_schemas : Literal["exactly", "names", "lax"]
            How to check the schemas of the before and after values.
            - "exactly": both before and after must have the same columns and types.
            - "names": both before and after must have the same columns, but types can differ.
            - "lax": no schema checking, just that there is at least one of 'before' or 'after' in each column.

        Returns
        -------
        Updates
            An Updates object representing the changes.
        """  # noqa: E501
        if isinstance(before, ir.StructValue):
            before = {col: before[col] for col in before.type().names}
        if isinstance(after, ir.StructValue):
            after = {col: after[col] for col in after.type().names}

        all_columns = set(before.keys()) | set(after.keys())

        def make_diff_col(col: str) -> ir.StructColumn:
            d = {}
            if col in before:
                d["before"] = before[col]
            if col in after:
                d["after"] = after[col]
            return ibis.struct(d).name(col)

        diff_table = _util.select(*[make_diff_col(c) for c in all_columns])
        return cls(diff_table, check_schemas=check_schemas)

    def before_values(self) -> dict[str, ir.Column]:
        """The values before the changes."""
        return {
            col: self[col].before
            for col in self.columns
            if "before" in self[col].type().names
        }

    def after_values(self) -> dict[str, ir.Column]:
        """The values after the changes."""
        return {
            col: self[col].after
            for col in self.columns
            if "after" in self[col].type().names
        }

    def before(self) -> ibis.Table:
        """The table before the changes."""
        return self.select(**self.before_values())

    def after(self) -> ibis.Table:
        """The table after the changes."""
        return self.select(**self.after_values())

    def is_changed(self, column: str, /) -> ibis.ir.BooleanColumn:
        """Is column.before different from column.after? Never returns NULL."""
        (val,) = self.bind(column)
        return is_changed(val)

    def filter(self, *args, **kwargs):
        return self.__class__(
            self._t.filter(*args, **kwargs), check_schemas=self.check_schemas
        )

    def cache(self):
        return self.__class__(self._t.cache(), check_schemas=self.check_schemas)

    def apply_to(
        self,
        t: ibis.Table,
        /,
        *,
        defaults: None | Any = _util.NOT_SET,
    ) -> ibis.Table:
        """Return the input table with these updates applied to it.

        Parameters
        ----------
        t
            The table to apply the updates to.
        defaults
            If the after table has more columns than the before table, you must provide defaults.
            This is because there might be some rows in `t` that are not touched by this
            Updates. We need to know what to put in those columns for the untouched rows.
            This accepts anything that `ibis.Table.mutate()` accepts.

            If None, as convenience, we will use `null` as the default for all new columns.
            If _util.NOT_SET, an error is raised if the after table has more columns than the before table.
        """  # noqa: E501
        _util.check_schemas_equal(t, self.before())

        t = t.difference(self.before(), distinct=False)

        if self.before().schema() == self.after().schema():
            # easy, we don't have to worry about adding defaults
            return t.union(self.after(), distinct=False)

        missing_cols = [
            c for c in self.after().schema() if c not in self.before().schema()
        ]
        if missing_cols:
            if defaults is _util.NOT_SET:
                raise ValueError(
                    "If the after table has more columns than the before table, you must provide defaults"  # noqa: E501
                )
            if defaults is None:
                defaults = {c: None for c in missing_cols}

            defaults_cols = t.select(defaults).columns
            already_there = [c for c in defaults_cols if c in t.columns]
            if already_there:
                raise ValueError(
                    f"default value {already_there} already exist in the input table"
                )

            t = t.mutate(defaults)

        t = t.select(self.after().columns)
        t = t.union(self.after(), distinct=False)
        return t


def is_changed(val: ibis.Value, /) -> ibis.ir.BooleanColumn:
    """Is val.before different from val.after? Never returns NULL."""
    return ibis.or_(
        (val.before != val.after).fill_null(False),
        val.before.isnull() != val.after.isnull(),
    )



from __future__ import annotations

import ibis
from ibis.expr import datatypes as dt
import pandas as pd
import pytest

from mismo import arrays

mark_dtype = pytest.mark.parametrize("dtype", [dt.Int16(), dt.Float32()])


@mark_dtype
@pytest.mark.parametrize(
    "inp,exp",
    [
        pytest.param([1, 2, 3, None], 1, id="normal"),
        pytest.param([None, None], None, id="nulls"),
        pytest.param([], None, id="empty"),
        pytest.param(None, None, id="null"),
    ],
)
def test_array_min(backend, dtype, inp, exp):
    a = ibis.literal(inp).cast(dt.Array(value_type=dtype))
    result = arrays.array_min(a)
    assert result.type() == dtype
    r = result.execute()
    if pd.isna(r):
        r = None
    assert r == exp


@mark_dtype
@pytest.mark.parametrize(
    "inp,exp",
    [
        pytest.param([1, 2, 3, None], 3, id="normal"),
        pytest.param([None, None], None, id="nulls"),
        pytest.param([], None, id="empty"),
        pytest.param(None, None, id="null"),
    ],
)
def test_array_max(backend, dtype, inp, exp):
    a = ibis.literal(inp).cast(dt.Array(value_type=dtype))
    result = arrays.array_max(a)
    assert result.type() == dtype
    r = result.execute()
    if pd.isna(r):
        r = None
    assert r == exp


@pytest.mark.parametrize(
    "fn",
    [
        pytest.param(arrays.array_median, id="median"),
        pytest.param(arrays.array_mean, id="mean"),
    ],
)
@pytest.mark.parametrize(
    "inp,type,exp",
    [
        # duckdb maybe is inconsistent with how it takes the median of
        # even-length arrays. Skip testing that for now.
        # SELECT MEDIAN(x) from (SELECT unnest([0.0, 2.0]) as x); -> 0.0
        # SELECT MEDIAN(x) from (SELECT unnest([0, 2]) as x); -> 1.0
        # This sounds to me like the exact inverse of the docs,
        # which say that ordinal values get floored and quantitative values get meaned.
        # https://discord.com/channels/909674491309850675/921073327009853451/1278856039164411916
        pytest.param([0, 1, 2], "int", 1.0, id="happy"),
        pytest.param([0, 2], "int", 1.0, id="split"),
        pytest.param([0, 1], "int", 0.5, id="frac"),
        pytest.param([0.0, 1.0, 2.0], "float", 1.0, id="frac_float"),
        pytest.param([0, 1, None, 2], "int", 1.0, id="with_null"),
        pytest.param([], "int", None, id="empty"),
        pytest.param(None, "int", None, id="null"),
        pytest.param([None, None], "int", None, id="all_null"),
    ],
)
def test_array_mean_median(backend, fn, inp, type, exp):
    a = ibis.literal(inp, type=f"array<{type}>")
    result = fn(a)
    assert result.type().is_floating()
    r = result.execute()
    if pd.isna(r):
        r = None
    assert r == exp


@pytest.mark.parametrize(
    "inp,exp",
    [
        pytest.param([True, False, True, None], False, id="mixed"),
        pytest.param([True, True], True, id="true"),
        pytest.param([True, True, None], True, id="true_null"),
        pytest.param([False, False], False, id="false"),
        pytest.param([False, False, None], False, id="false_null"),
        pytest.param([None, None], None, id="nulls"),
        pytest.param([], None, id="empty"),
        pytest.param(None, None, id="null"),
    ],
)
def test_array_all(backend, inp, exp):
    a = ibis.literal(inp, type="array<boolean>")
    result = arrays.array_all(a)
    assert result.type() == dt.Boolean()
    r = result.execute()
    assert r == exp


@pytest.mark.parametrize(
    "inp,exp",
    [
        pytest.param([True, False, True, None], True, id="mixed"),
        pytest.param([True, True], True, id="true"),
        pytest.param([True, True, None], True, id="true_null"),
        pytest.param([False, False], False, id="false"),
        pytest.param([False, False, None], False, id="false_null"),
        pytest.param([None, None], None, id="nulls"),
        pytest.param([], None, id="empty"),
        pytest.param(None, None, id="null"),
    ],
)
def test_array_any(backend, inp, exp):
    a = ibis.literal(inp, type="array<boolean>")
    result = arrays.array_any(a)
    assert result.type() == dt.Boolean()
    r = result.execute()
    assert r == exp


@pytest.mark.parametrize(
    "key,exp",
    [
        pytest.param(None, ["a", "a", "b", "c"], id="none"),
        pytest.param((lambda x: x.date), ["b", "a", "c", "a"], id="lambda"),
        pytest.param(
            (lambda x: [x.date, x.email]), ["b", "a", "c", "a"], id="lambda_multiple"
        ),
        pytest.param(ibis._.date, ["b", "a", "c", "a"], id="deferred"),
        pytest.param(
            [ibis._.date, ibis._.email], ["b", "a", "c", "a"], id="date_then_email"
        ),
        pytest.param(
            [ibis._.email, ibis._.date], ["a", "a", "b", "c"], id="email_then_date"
        ),
    ],
)
def test_array_sort(key, exp):
    emails = ibis.array(
        [
            ibis.struct({"email": "b", "date": 1}),
            ibis.struct({"email": "c", "date": 3}),
            ibis.struct({"email": "a", "date": 2}),
            ibis.struct({"email": "a", "date": 4}),
        ]
    )
    result = arrays.array_sort(emails, key=key)
    emails = result.map(lambda x: x.email)
    assert emails.execute() == exp


@pytest.mark.parametrize(
    "other",
    [
        pytest.param([2], id="normal"),
        pytest.param([1, 2, 3], id="remove_nothing"),
        pytest.param([], id="remove_all"),
    ],
)
@pytest.mark.parametrize(
    "inp",
    [
        pytest.param([1, 2, 3], id="normal"),
        pytest.param([1, 2, 2], id="duplicates"),
        pytest.param([2, None, 2], id="duplicates_null"),
        pytest.param([1], id="single"),
        pytest.param([1, None, 2], id="mixed"),
        pytest.param([None, 1], id="start_null"),
        pytest.param([1, None], id="end_null"),
        pytest.param([], id="empty"),
        pytest.param(None, id="null"),
    ],
)
def test_array_filter_isin_other(table_factory, column_factory, inp, other):
    # We make 5 rows (admittedly, all with the same data)
    # because the implementation does some unnesting logic
    # so we want to test that the output rows line up with the input rows
    pairs = [{"inp": inp} for _ in range(5)]
    t = table_factory(pairs, schema={"inp": "array<int32>"})
    other_col = column_factory(other, type="int32")
    filtered = arrays.array_filter_isin_other(
        t, "inp", other_col, result_format="filtered"
    )
    df = filtered.execute()

    def baseline(inp):
        if inp is None:
            return None
        return [x for x in inp if x in other or pd.isna(x)]

    assert len(df) == len(pairs)
    assert df.filtered.tolist() == [baseline(x) for x in df.inp.tolist()]



"""Utils for cleaning and comparing email addresses."""

from __future__ import annotations

from mismo.lib.email._core import EmailMatchLevel as EmailMatchLevel
from mismo.lib.email._core import EmailsDimension as EmailsDimension
from mismo.lib.email._core import ParsedEmail as ParsedEmail
from mismo.lib.email._core import clean_email as clean_email
from mismo.lib.email._core import match_level as match_level



"""
Term Frequency utils, eg matching on "John Smith" has less signal than "Arun di Suvero".
"""

from __future__ import annotations

from mismo.tf._filterer import AmbiguousHaystackFilterer as AmbiguousHaystackFilterer
from mismo.tf._filterer import RareLookupFilterer as RareLookupFilterer
from mismo.tf._tf import ColumnStats as ColumnStats
from mismo.tf._tf import StatsTable as StatsTable
from mismo.tf._tf import TermFrequencyModel as TermFrequencyModel



from __future__ import annotations

from pathlib import Path

import ibis
from ibis.expr import types as ir

from mismo import _util

__all__ = [
    "load_febrl1",
    "load_febrl2",
    "load_febrl3",
    "load_patents",
    "load_rldata500",
    "load_rldata10000",
]

_DATASETS_DIR = Path(__file__).parent / "_data/_datasets"


def _wrap_febrl(loader_name: str) -> tuple[ir.Table, ir.Table]:
    with _util.optional_import("recordlinkage"):
        from recordlinkage import datasets as rlds

    loader = getattr(rlds, loader_name)
    pdf, links_multi_index = loader(return_links=True)
    pdf = pdf.reset_index(drop=False)
    schema = {
        "rec_id": "str",
        "given_name": "str",
        "surname": "str",
        "street_number": "str",  # keep as string for leading 0s
        "address_1": "str",
        "address_2": "str",
        "suburb": "str",
        "postcode": "str",  # keep as string for leading 0s
        "state": "str",
        "soc_sec_id": "int32",  # 7 digits long, never null, no leading 0s
        "date_of_birth": "str",  # contains some BS dates like 19371233
    }
    t = ibis.memtable(pdf)
    t = t.cast(schema)
    t = t.rename(record_id="rec_id")
    t = t.order_by("record_id")
    t = t.cache()

    links_df = links_multi_index.to_frame(
        index=False, name=["record_id_l", "record_id_r"]
    )
    links = ibis.memtable(links_df)
    links = links.order_by(["record_id_l", "record_id_r"])
    links = links.cache()
    return (t, links)


def load_febrl1() -> tuple[ir.Table, ir.Table]:
    """Load the FEBRL 1 dataset.

    The Freely Extensible Biomedical Record Linkage (Febrl) package is distributed
    with a dataset generator and four datasets generated with the generator.
    """
    return _wrap_febrl("load_febrl1")


def load_febrl2() -> tuple[ir.Table, ir.Table]:
    """Load the FEBRL 2 dataset.

    The Freely Extensible Biomedical Record Linkage (Febrl) package is distributed
    with a dataset generator and four datasets generated with the generator.
    """
    return _wrap_febrl("load_febrl2")


def load_febrl3() -> tuple[ir.Table, ir.Table]:
    """Load the FEBRL 3 dataset.

    The Freely Extensible Biomedical Record Linkage (Febrl) package is distributed
    with a dataset generator and four datasets generated with the generator.
    """
    return _wrap_febrl("load_febrl3")


# Don't bother wrapping load_febrl4 because it has a different API,
# could add that later if it's needed.


def load_patents(backend: ibis.BaseBackend | None = None) -> ir.Table:
    """Load the PATSTAT dataset

    This represents a dataset of patents, and the task is to determine which
    patents came from the same inventor.

    This comes from
    [the Dedupe Patent Example](https://github.com/dedupeio/dedupe-examples/tree/master/patent_example).

    Returns
    -------
    Table
        An Ibis Table with the following schema:

        - record_id: int64
          A unique ID for each row in the table.
        - label_true: int64
          The manually labeled, true ID of the inventor.
        - name_true: str
          The manually labeled, true name of the inventor.
        - name: str
          The raw name on the patent.
        - latitude: float64
          Geocoded from the inventor's address. 0.0 indicates no address was found
        - longitude: float64
        - coauthor: str
          A list of coauthors on the patent, separated by "**"
        - class_: str
          A list of 4-character IPC technical codes, separated by "**"

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> load_patents().order_by("record_id").head()  # doctest: +SKIP
    ┏━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ record_id ┃ label_true ┃ name_true            ┃ name                                             ┃ latitude ┃ longitude ┃ coauthors                                       ┃ classes                                         ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
    │ int64     │ int64      │ string               │ string                                           │ float64  │ float64   │ string                                          │ string                                          │
    ├───────────┼────────────┼──────────────────────┼──────────────────────────────────────────────────┼──────────┼───────────┼─────────────────────────────────────────────────┼─────────────────────────────────────────────────┤
    │      2909 │     402600 │ AGILENT TECHNOLOGIES │ * AGILENT TECHNOLOGIES, INC.                     │     0.00 │  0.000000 │ KONINK PHILIPS ELECTRONICS N V**DAVID E  SNYDE… │ A61N**A61B                                      │
    │      3574 │     569309 │ AKZO NOBEL           │ * AKZO NOBEL N.V.                                │     0.00 │  0.000000 │ TSJERK  HOEKSTRA**ANDRESS K  JOHNSON**TERESA M… │ G01N**B01L**C11D**G02F**F16L                    │
    │      3575 │     569309 │ AKZO NOBEL           │ * AKZO NOBEL NV                                  │     0.00 │  0.000000 │ WILLIAM JOHN ERNEST  PARR**HANS  OSKARSSON**MA… │ C09K**F17D**B01F**C23F                          │
    │      3779 │     656303 │ ALCATEL              │ * ALCATEL N.V.                                   │    52.35 │  4.916667 │ GUENTER  KOCHSMEIER**ZBIGNIEW  WIEGOLASKI**EVA… │ G02B**G04G**H02G**G06F                          │
    │      3780 │     656303 │ ALCATEL              │ * ALCATEL N.V.                                   │    52.35 │  4.916667 │ ZILAN  MANFRED**JOSIANE  RAMOS**DUANE LYNN  MO… │ H03G**B05D**H04L**H04B**C03B**C03C**G02B**H01B  │
    └───────────┴────────────┴──────────────────────┴──────────────────────────────────────────────────┴──────────┴───────────┴─────────────────────────────────────────────────┴─────────────────────────────────────────────────┘
    """  # noqa E501
    path = _DATASETS_DIR / "patstat/patents.csv"
    if backend is None:
        backend = ibis
    # In order to guarantee row order, could either use
    # parallel=False kwarg, but I'd rather just have them sorted
    # by record_id
    return backend.read_csv(path).order_by("record_id").cache()


def rldata_schema() -> dict:
    return {
        "record_id": "int64",
        "label_true": "int64",
        "fname_c1": "string",
        "fname_c2": "string",
        "lname_c1": "string",
        "lname_c2": "string",
        "by": "int64",
        "bm": "int64",
        "bd": "int64",
    }


def load_rldata500(backend: ibis.BaseBackend | None = None) -> ir.Table:
    """Synthetic personal information dataset with 500 rows

    This is a synthetic dataset with noisy names and dates of birth, with the task being
    to determine which rows represent the same person. 10% of the records are duplicates
    of existing ones, and the level of noise is low. The dataset can be deduplicated with 90%+
    precision and recall using simple linkage rules. It is often used as a
    sanity check for computational efficiency and disambiguation accuracy.

    This comes from the
    [RecordLinkage R package](https://cran.r-project.org/web/packages/RecordLinkage/index.html)
    and was generated using the data generation component of
    [Febrl (Freely Extensible Biomedical Record Linkage)](https://sourceforge.net/projects/febrl/).

    Returns
    -------
    Table
        An Ibis Table with the following schema:

        - record_id: int64
          A unique ID for each row in the table.
        - label_true: int64
          The manually labeled, true ID of the inventor.
        - fname_c1: string
          First component of the first name.
        - fname_c2: string
          Second component of the first name (mostly NULL values)
        - lname_c1: string
          First component of the last name.
        - lname_c2: string
          Second component of the last name (mostly NULL values).
        - by: int64
          Birth year
        - bm: int64
          Birth month
        - bd: int64
          Birth day

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> load_rldata500().head()  # doctest: +SKIP
    ┏━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━┳━━━━━━━┳━━━━━━━┓
    ┃ record_id ┃ label_true ┃ fname_c1 ┃ fname_c2 ┃ lname_c1 ┃ lname_c2 ┃ by    ┃ bm    ┃ bd    ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━╇━━━━━━━╇━━━━━━━┩
    │ int64     │ int64      │ string   │ string   │ string   │ string   │ int64 │ int64 │ int64 │
    ├───────────┼────────────┼──────────┼──────────┼──────────┼──────────┼───────┼───────┼───────┤
    │         0 │         34 │ CARSTEN  │ NULL     │ MEIER    │ NULL     │  1949 │     7 │    22 │
    │         1 │         51 │ GERD     │ NULL     │ BAUER    │ NULL     │  1968 │     7 │    27 │
    │         2 │        115 │ ROBERT   │ NULL     │ HARTMANN │ NULL     │  1930 │     4 │    30 │
    │         3 │        189 │ STEFAN   │ NULL     │ WOLFF    │ NULL     │  1957 │     9 │     2 │
    │         4 │         72 │ RALF     │ NULL     │ KRUEGER  │ NULL     │  1966 │     1 │    13 │
    └───────────┴────────────┴──────────┴──────────┴──────────┴──────────┴───────┴───────┴───────┘
    """  # noqa: E501
    path = _DATASETS_DIR / "rldata/RLdata500.csv"
    schema = rldata_schema()
    if backend is None:
        backend = ibis
    return (
        backend.read_csv(path)
        .cast(schema)
        .select(*schema.keys())
        .order_by("record_id")
        .cache()
    )


def load_rldata10000(backend: ibis.BaseBackend | None = None) -> ir.Table:
    """Synthetic personal information dataset with 10000 rows

    This is a synthetic dataset with noisy names and dates of birth, with the task being
    to determine which rows represent the same person. 10% of the records are duplicates
    of existing ones, and the level of noise is low. The dataset can be deduplicated with 90%+
    precision and recall using simple linkage rules. It is often used as a
    sanity check for computational efficiency and disambiguation accuracy.

    This comes from the
    [RecordLinkage R package](https://cran.r-project.org/web/packages/RecordLinkage/index.html)
    and was generated using the data generation component of
    [Febrl (Freely Extensible Biomedical Record Linkage)](https://sourceforge.net/projects/febrl/).

    Returns
    -------
    Table
        An Ibis Table with the following schema:

        - record_id: int64
          A unique ID for each row in the table.
        - label_true: int64
          The manually labeled, true ID of the inventor.
        - fname_c1: string
          First component of the first name.
        - fname_c2: string
          Second component of the first name (mostly NULL values)
        - lname_c1: string
          First component of the last name.
        - lname_c2: string
          Second component of the last name (mostly NULL values).
        - by: int64
          Birth year
        - bm: int64
          Birth month
        - bd: int64
          Birth day

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> load_rldata10000().head()  # doctest: +SKIP
    ┏━━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━━━━━━┳━━━━━━━━━━┳━━━━━━━┳━━━━━━━┳━━━━━━━┓
    ┃ record_id ┃ label_true ┃ fname_c1 ┃ fname_c2 ┃ lname_c1   ┃ lname_c2 ┃ by    ┃ bm    ┃ bd    ┃
    ┡━━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━━━━━━╇━━━━━━━━━━╇━━━━━━━╇━━━━━━━╇━━━━━━━┩
    │ int64     │ int64      │ string   │ string   │ string     │ string   │ int64 │ int64 │ int64 │
    ├───────────┼────────────┼──────────┼──────────┼────────────┼──────────┼───────┼───────┼───────┤
    │         0 │       3606 │ FRANK    │ NULL     │ MUELLER    │ NULL     │  1967 │     9 │    27 │
    │         1 │       2560 │ MARTIN   │ NULL     │ SCHWARZ    │ NULL     │  1967 │     2 │    17 │
    │         2 │       3892 │ HERBERT  │ NULL     │ ZIMMERMANN │ NULL     │  1961 │    11 │     6 │
    │         3 │        329 │ HANS     │ NULL     │ SCHMITT    │ NULL     │  1945 │     8 │    14 │
    │         4 │       1994 │ UWE      │ NULL     │ KELLER     │ NULL     │  2000 │     7 │     5 │
    └───────────┴────────────┴──────────┴──────────┴────────────┴──────────┴───────┴───────┴───────┘
    """  # noqa: E501
    path = _DATASETS_DIR / "rldata/RLdata10000.csv"
    schema = rldata_schema()
    if backend is None:
        backend = ibis
    return (
        backend.read_csv(path)
        .cast(schema)
        .select(*schema.keys())
        .order_by("record_id")
        .cache()
    )
